# Copyright 2026, Jiri Eischmann <jiri@eischmann.cz> and the meshy contributors
# SPDX-License-Identifier: GPL-3.0-or-later

"""Chat conversation view - shown in the content panel."""

import gi

gi.require_version("Gtk", "4.0")
gi.require_version("Adw", "1")
gi.require_version("Shumate", "1.0")
from gi.repository import Adw, Gdk, Gio, GLib, GObject, Gtk, Pango
from meshy.models import Contact, Message, MessageStatus
from meshy.protocol import max_dm_text_bytes
from meshy.utils import linkify, parse_shared_contact, parse_shared_location
from meshy.views import _format_date_text
from meshy.views.chat_helpers import (
    CONTACT_TYPE_ICONS,
    ensure_loc_map,
    open_location_external,
    share_contact_dialog,
    share_location,
    share_location_map,
    show_location_map,
)
from meshy.views.chat_mixin import ChatScrollMixin

_BATCH_SIZE = 30


class ChatItem(GObject.Object):
    """Model object for one row in the chat ListView."""

    KIND_MESSAGE = 0
    KIND_UNREAD_DIVIDER = 1
    KIND_DATE_DIVIDER = 2

    def __init__(self, kind=0, message=None, date_text=""):
        super().__init__()
        self.kind = kind
        self.message = message
        self.date_text = date_text
        self._status_val = int(message.status) if message else int(MessageStatus.PENDING)
        self.status_detail = ""

    @GObject.Property(type=int)
    def status(self):
        return self._status_val

    @status.setter
    def status(self, value):
        self._status_val = value


@Gtk.Template(resource_path="/page/codeberg/sesivany/Meshy/ui/chat-view.ui")
class ChatView(ChatScrollMixin, Gtk.Box):
    __gtype_name__ = "MeshyChatView"

    _stack = Gtk.Template.Child("stack")
    _scroll = Gtk.Template.Child("scroll")
    _list_view = Gtk.Template.Child("list_view")
    _scroll_down_revealer = Gtk.Template.Child("scroll_down_revealer")
    _unread_btn_revealer = Gtk.Template.Child("unread_btn_revealer")
    _char_label = Gtk.Template.Child("char_label")
    _emoji_btn = Gtk.Template.Child("emoji_btn")
    _text_entry = Gtk.Template.Child("text_entry")
    _send_btn = Gtk.Template.Child("send_btn")

    def __init__(self, window, **kwargs):
        super().__init__(**kwargs)
        self._window = window
        self._contact: Contact | None = None
        self._item_by_msg_id: dict[str, ChatItem] = {}

        # Pango attributes for char counter
        attrs = Pango.AttrList()
        attrs.insert(Pango.attr_scale_new(0.85))
        self._char_label.set_attributes(attrs)

        # ListView model and factory
        self._store = Gio.ListStore(item_type=ChatItem)
        selection = Gtk.NoSelection(model=self._store)
        self._factory = Gtk.SignalListItemFactory()
        self._factory.connect("setup", self._on_factory_setup)
        self._factory.connect("bind", self._on_factory_bind)
        self._factory.connect("unbind", self._on_factory_unbind)
        self._list_view.set_model(selection)
        self._list_view.set_factory(self._factory)

        # Context menu for outgoing messages
        self._context_popover = Gtk.PopoverMenu()
        self._context_popover.set_parent(self._list_view)
        self._context_popover.set_has_arrow(False)
        self._context_popover.connect("closed", self._on_context_popover_closed)
        self._context_menu_open = False
        self._frozen_scroll_value = 0.0
        ctx_menu = Gio.Menu()
        ctx_menu.append(_("Send Again"), "msg.resend")
        ctx_menu.append(_("Delete"), "msg.delete")
        self._context_popover.set_menu_model(ctx_menu)
        self._context_item = None

        ctx_group = Gio.SimpleActionGroup()
        resend_a = Gio.SimpleAction.new("resend", None)
        resend_a.connect("activate", lambda a, p: self._resend_context_item())
        ctx_group.add_action(resend_a)
        delete_a = Gio.SimpleAction.new("delete", None)
        delete_a.connect("activate", lambda a, p: self._delete_context_item())
        ctx_group.add_action(delete_a)
        self._list_view.insert_action_group("msg", ctx_group)

        # Gesture controllers
        right_click = Gtk.GestureClick(button=3)
        right_click.set_propagation_phase(Gtk.PropagationPhase.CAPTURE)
        right_click.connect("pressed", self._on_list_right_click)
        self._list_view.add_controller(right_click)
        long_press = Gtk.GestureLongPress()
        long_press.set_touch_only(True)
        long_press.set_exclusive(True)
        long_press.connect("pressed", self._on_list_long_press)
        self._list_view.add_controller(long_press)

        self._is_sticky = True
        self._scroll_positions: dict[str, float] = {}
        self._restoring_scroll = False
        self._unread_pos = -1
        self._messages_offset = 0
        self._all_loaded = False
        self._loading_more = False

        # Signal connections
        self._scroll_down_revealer.get_child().connect("clicked", self._on_scroll_to_bottom_clicked)
        self._unread_btn_revealer.get_child().connect("clicked", self._on_scroll_to_unread_clicked)
        self._emoji_btn.connect("clicked", self._on_emoji_clicked)
        self._text_entry.connect("activate", self._on_send)
        self._text_entry.connect("changed", self._on_text_changed)
        self._send_btn.connect("clicked", self._on_send)

        # Send route override menu (right-click / long-press on send button)
        send_menu = Gio.Menu()
        send_menu.append(_("Send Automatically"), "send.auto")
        send_menu.append(_("Send Direct"), "send.direct")
        send_menu.append(_("Send Flood"), "send.flood")
        self._send_popover = Gtk.PopoverMenu(menu_model=send_menu)
        self._send_popover.set_parent(self._send_btn)
        self._send_popover.set_has_arrow(False)

        send_group = Gio.SimpleActionGroup()
        for name, override in [("auto", None), ("direct", "direct"), ("flood", "flood")]:
            action = Gio.SimpleAction.new(name, None)
            action.connect("activate", lambda _a, _p, ov=override: self._on_send_with_override(ov))
            send_group.add_action(action)
        self._send_btn.insert_action_group("send", send_group)

        send_right_click = Gtk.GestureClick(button=3)
        send_right_click.connect("pressed", lambda _g, _n, x, y: self._send_popover.popup())
        self._send_btn.add_controller(send_right_click)
        send_long_press = Gtk.GestureLongPress()
        send_long_press.connect("pressed", lambda _g, x, y: self._send_popover.popup())
        self._send_btn.add_controller(send_long_press)

        # Extra context menu on text entry
        extra_menu = Gio.Menu()
        extra_menu.append(_("Share Contact"), "entry.share-contact")
        extra_menu.append(_("Share Location"), "entry.share-location")
        extra_menu.append(_("Share Location from Map"), "entry.share-location-map")
        self._text_entry.set_extra_menu(extra_menu)
        entry_group = Gio.SimpleActionGroup()
        share_a = Gio.SimpleAction.new("share-contact", None)
        share_a.connect("activate", self._on_share_contact)
        entry_group.add_action(share_a)
        loc_a = Gio.SimpleAction.new("share-location", None)
        loc_a.connect("activate", self._on_share_location)
        entry_group.add_action(loc_a)
        loc_map_a = Gio.SimpleAction.new("share-location-map", None)
        loc_map_a.connect("activate", self._on_share_location_map)
        entry_group.add_action(loc_map_a)
        self._text_entry.insert_action_group("entry", entry_group)

        # Scroll state tracking
        adj = self._scroll.get_vadjustment()
        adj.connect("value-changed", self._on_scroll_value_changed)
        adj.connect("notify::upper", self._on_scroll_upper_changed)
        adj.connect("notify::page-size", self._on_scroll_upper_changed)
        self._scroll.connect("edge-overshot", self._on_edge_overshot)

        self._stack.set_visible_child_name("placeholder")

    # ─── Factory callbacks ───────────────────────────────────

    def _on_factory_setup(self, factory, list_item):
        """Create the widget tree for one list item (recycled)."""
        list_item.set_activatable(False)
        list_item.set_selectable(False)
        list_item.set_focusable(False)
        outer = Gtk.Box(orientation=Gtk.Orientation.VERTICAL)

        # ── Message bubble ──
        msg_box = Gtk.Box(
            orientation=Gtk.Orientation.VERTICAL,
            spacing=2,
            margin_top=4,
            margin_bottom=4,
            margin_start=12,
            margin_end=12,
        )

        # Sender name (for room messages)
        sender_lbl = Gtk.Label(xalign=0, max_width_chars=40)
        sender_lbl.add_css_class("caption")
        sender_lbl.add_css_class("accent")
        sender_lbl.set_margin_start(4)
        sender_lbl.set_visible(False)
        msg_box.append(sender_lbl)

        # Card frame with text
        text_lbl = Gtk.Label(
            wrap=True,
            wrap_mode=Pango.WrapMode.WORD_CHAR,
            xalign=0,
            selectable=True,
            max_width_chars=40,
        )
        text_lbl.set_margin_start(12)
        text_lbl.set_margin_end(12)
        text_lbl.set_margin_top(8)
        text_lbl.set_margin_bottom(8)

        frame = Gtk.Frame()
        frame.add_css_class("card")

        content_stack = Gtk.Stack()
        content_stack.set_transition_type(Gtk.StackTransitionType.NONE)
        content_stack.set_vhomogeneous(False)
        content_stack.set_hhomogeneous(False)
        content_stack.add_named(text_lbl, "text")

        sc_box = Gtk.Box(
            orientation=Gtk.Orientation.VERTICAL,
            spacing=6,
            margin_start=12,
            margin_end=12,
            margin_top=8,
            margin_bottom=8,
        )
        sc_header_lbl = Gtk.Label(xalign=0, wrap=True)
        sc_header_lbl.add_css_class("dim-label")
        sc_header_lbl.add_css_class("caption")
        sc_box.append(sc_header_lbl)

        sc_name_box = Gtk.Box(spacing=6)
        sc_icon = Gtk.Image.new_from_icon_name("avatar-default-symbolic")
        sc_name_lbl = Gtk.Label(xalign=0)
        attrs = Pango.AttrList()
        attrs.insert(Pango.attr_weight_new(Pango.Weight.BOLD))
        sc_name_lbl.set_attributes(attrs)
        sc_name_box.append(sc_icon)
        sc_name_box.append(sc_name_lbl)
        sc_box.append(sc_name_box)

        sc_key_lbl = Gtk.Label(xalign=0)
        sc_key_lbl.add_css_class("dim-label")
        sc_key_lbl.add_css_class("caption")
        sc_box.append(sc_key_lbl)

        sc_add_btn = Gtk.Button(label=_("Add Contact"))
        sc_add_btn.add_css_class("suggested-action")
        sc_add_btn.set_halign(Gtk.Align.START)
        sc_add_btn.set_margin_top(4)
        sc_box.append(sc_add_btn)

        content_stack.add_named(sc_box, "contact")

        # ── Shared location card ──
        loc_box = Gtk.Box(
            orientation=Gtk.Orientation.VERTICAL,
            spacing=6,
            margin_start=12,
            margin_end=12,
            margin_top=8,
            margin_bottom=8,
        )
        loc_header_lbl = Gtk.Label(xalign=0, wrap=True)
        loc_header_lbl.add_css_class("dim-label")
        loc_header_lbl.add_css_class("caption")
        loc_box.append(loc_header_lbl)

        loc_coords_lbl = Gtk.Label(xalign=0, selectable=True)
        loc_coords_lbl.add_css_class("caption")
        loc_box.append(loc_coords_lbl)

        loc_map_frame = Gtk.Frame()
        loc_map_frame.add_css_class("card")
        loc_map_frame.set_overflow(Gtk.Overflow.HIDDEN)
        loc_map_frame.set_cursor(Gdk.Cursor.new_from_name("pointer"))
        loc_map_gesture = Gtk.GestureClick()
        loc_map_frame.add_controller(loc_map_gesture)
        loc_box.append(loc_map_frame)

        loc_open_btn = Gtk.Button(label=_("Open in Maps App"), margin_top=4)
        loc_box.append(loc_open_btn)

        content_stack.add_named(loc_box, "location")

        frame.set_child(content_stack)
        msg_box.append(frame)

        # Meta line: time + status
        meta_box = Gtk.Box(spacing=4)
        time_lbl = Gtk.Label()
        time_lbl.add_css_class("dim-label")
        time_lbl.add_css_class("caption")
        meta_box.append(time_lbl)

        status_lbl = Gtk.Label()
        status_lbl.add_css_class("caption")
        status_lbl.set_visible(False)
        meta_box.append(status_lbl)

        msg_box.append(meta_box)
        outer.append(msg_box)

        # ── Divider ──
        div_box = Gtk.Box(
            spacing=8,
            margin_start=12,
            margin_end=12,
            margin_top=8,
            margin_bottom=8,
        )
        line_left = Gtk.Separator(hexpand=True, valign=Gtk.Align.CENTER)
        div_box.append(line_left)
        div_lbl = Gtk.Label()
        div_box.append(div_lbl)
        line_right = Gtk.Separator(hexpand=True, valign=Gtk.Align.CENTER)
        div_box.append(line_right)
        div_box.set_visible(False)
        outer.append(div_box)

        # Store widget references
        outer._msg_box = msg_box
        outer._sender_lbl = sender_lbl
        outer._text_lbl = text_lbl
        outer._frame = frame
        outer._time_lbl = time_lbl
        outer._status_lbl = status_lbl
        outer._meta_box = meta_box
        outer._div_box = div_box
        outer._div_lbl = div_lbl
        outer._div_line_left = line_left
        outer._div_line_right = line_right
        outer._content_stack = content_stack
        outer._sc_header_lbl = sc_header_lbl
        outer._sc_icon = sc_icon
        outer._sc_name_lbl = sc_name_lbl
        outer._sc_key_lbl = sc_key_lbl
        outer._sc_add_btn = sc_add_btn
        outer._sc_add_handler_id = None
        outer._loc_header_lbl = loc_header_lbl
        outer._loc_coords_lbl = loc_coords_lbl
        outer._loc_map_frame = loc_map_frame
        outer._loc_open_btn = loc_open_btn
        outer._loc_map_gesture = loc_map_gesture
        outer._loc_map_click_handler = None
        outer._loc_open_handler = None
        outer._loc_map_widget = None
        outer._loc_viewport = None
        outer._loc_marker = None
        outer._chat_item = None
        outer._status_handler_id = None

        # Wrap in per-item Clamp for width limiting
        clamp = Adw.Clamp(maximum_size=600, child=outer)
        clamp._outer = outer
        list_item.set_child(clamp)

    def _on_factory_bind(self, factory, list_item):
        """Populate widgets from the model item."""
        item = list_item.get_item()
        clamp = list_item.get_child()
        outer = clamp._outer
        outer._chat_item = item

        if item.kind == ChatItem.KIND_MESSAGE:
            msg = item.message
            outer._msg_box.set_visible(True)
            outer._div_box.set_visible(False)

            # Alignment
            if msg.is_outgoing:
                outer._msg_box.set_halign(Gtk.Align.END)
                outer._meta_box.set_halign(Gtk.Align.END)
            else:
                outer._msg_box.set_halign(Gtk.Align.START)
                outer._meta_box.set_halign(Gtk.Align.START)

            # Sender name (room messages)
            if msg.room_sender_name and not msg.is_outgoing:
                outer._sender_lbl.set_label(msg.room_sender_name)
                outer._sender_lbl.set_visible(True)
            else:
                outer._sender_lbl.set_visible(False)

            # Text / shared contact
            sc = parse_shared_contact(msg.text)
            if sc:
                pub_key_hex, contact_type, contact_name = sc
                outer._content_stack.set_visible_child_name("contact")
                if msg.is_outgoing:
                    outer._sc_header_lbl.set_label(_("You shared a contact:"))
                else:
                    sender = self._contact.name if self._contact else "?"
                    outer._sc_header_lbl.set_label(
                        _("{name} shared a contact with you:").format(name=sender)
                    )
                outer._sc_icon.set_from_icon_name(
                    CONTACT_TYPE_ICONS.get(contact_type, "avatar-default-symbolic")
                )
                outer._sc_name_lbl.set_label(contact_name)
                outer._sc_key_lbl.set_label(f"{pub_key_hex[:8]}...{pub_key_hex[-8:]}")
                if msg.is_outgoing:
                    outer._sc_add_btn.set_visible(False)
                else:
                    outer._sc_add_btn.set_visible(True)
                    is_known = self._window.is_contact_known(pub_key_hex)
                    outer._sc_add_btn.set_sensitive(not is_known)
                    outer._sc_add_btn.set_label(
                        _("Already added") if is_known else _("Add Contact")
                    )
                    if outer._sc_add_handler_id is not None:
                        outer._sc_add_btn.disconnect(outer._sc_add_handler_id)
                    outer._sc_add_handler_id = outer._sc_add_btn.connect(
                        "clicked",
                        self._on_shared_contact_add,
                        pub_key_hex,
                        contact_type,
                        contact_name,
                    )
            else:
                loc = parse_shared_location(msg.text)
                if loc:
                    lat, lon = loc
                    outer._content_stack.set_visible_child_name("location")
                    if msg.is_outgoing:
                        outer._loc_header_lbl.set_label(_("You shared a location:"))
                    else:
                        sender = self._contact.name if self._contact else "?"
                        outer._loc_header_lbl.set_label(
                            _("{name} shared a location:").format(name=sender)
                        )
                    outer._loc_coords_lbl.set_label(f"{lat:.6f}, {lon:.6f}")
                    ensure_loc_map(outer, lat, lon)
                    outer._loc_map_click_handler = outer._loc_map_gesture.connect(
                        "released",
                        lambda g, n, x, y, la=lat, lo=lon: show_location_map(self._window, la, lo),
                    )
                    outer._loc_open_handler = outer._loc_open_btn.connect(
                        "clicked", open_location_external, lat, lon
                    )
                else:
                    outer._content_stack.set_visible_child_name("text")
                    outer._text_lbl.set_markup(linkify(msg.text))

            # Frame style
            if msg.is_outgoing:
                outer._frame.add_css_class("outgoing-message")
            else:
                outer._frame.remove_css_class("outgoing-message")

            # Timestamp
            outer._time_lbl.set_label(msg.timestamp.strftime("%H:%M"))

            # Status (outgoing only)
            if msg.is_outgoing:
                outer._status_lbl.set_visible(True)
                self._apply_status(
                    outer._status_lbl,
                    MessageStatus(item.status),
                    item.status_detail,
                )
                hid = item.connect(
                    "notify::status",
                    lambda obj, pspec, w=outer: self._apply_status(
                        w._status_lbl,
                        MessageStatus(obj.status),
                        obj.status_detail,
                    ),
                )
                outer._status_handler_id = hid
            else:
                outer._status_lbl.set_visible(False)

        elif item.kind == ChatItem.KIND_DATE_DIVIDER:
            outer._msg_box.set_visible(False)
            outer._div_box.set_visible(True)
            outer._div_lbl.set_label(item.date_text)
            # Date divider styling
            outer._div_lbl.remove_css_class("unread-divider")
            outer._div_lbl.add_css_class("dim-label")
            outer._div_line_left.remove_css_class("unread-divider-line")
            outer._div_line_left.add_css_class("date-divider-line")
            outer._div_line_right.remove_css_class("unread-divider-line")
            outer._div_line_right.add_css_class("date-divider-line")

        elif item.kind == ChatItem.KIND_UNREAD_DIVIDER:
            outer._msg_box.set_visible(False)
            outer._div_box.set_visible(True)
            outer._div_lbl.set_label(_("New Messages"))
            # Unread divider styling
            outer._div_lbl.remove_css_class("dim-label")
            outer._div_lbl.add_css_class("unread-divider")
            outer._div_line_left.remove_css_class("date-divider-line")
            outer._div_line_left.add_css_class("unread-divider-line")
            outer._div_line_right.remove_css_class("date-divider-line")
            outer._div_line_right.add_css_class("unread-divider-line")

    def _on_factory_unbind(self, factory, list_item):
        """Clean up signal connections when the widget is recycled."""
        clamp = list_item.get_child()
        outer = clamp._outer
        if outer._status_handler_id is not None:
            outer._chat_item.disconnect(outer._status_handler_id)
            outer._status_handler_id = None
        if outer._sc_add_handler_id is not None:
            outer._sc_add_btn.disconnect(outer._sc_add_handler_id)
            outer._sc_add_handler_id = None
        if outer._loc_map_click_handler is not None:
            outer._loc_map_gesture.disconnect(outer._loc_map_click_handler)
            outer._loc_map_click_handler = None
        if outer._loc_open_handler is not None:
            outer._loc_open_btn.disconnect(outer._loc_open_handler)
            outer._loc_open_handler = None
        outer._chat_item = None

    # ─── Shared contact helpers ─────────────────────────────

    def _on_shared_contact_add(self, btn, pub_key_hex, contact_type, name):
        self._window.add_contact(bytes.fromhex(pub_key_hex), contact_type, name)
        btn.set_sensitive(False)
        btn.set_label(_("Already added"))

    # ─── Status helpers ──────────────────────────────────────

    @staticmethod
    def _apply_status(label, status, detail=""):
        symbol, css = {
            MessageStatus.PENDING: (_("Sending"), "dim-label"),
            MessageStatus.SENT: (_("Sending"), "dim-label"),
            MessageStatus.DELIVERED: ("\u2713", "success"),
            MessageStatus.FAILED: ("\u2717", "error"),
        }.get(status, ("...", "dim-label"))
        text = f"{symbol} {detail}".strip() if detail else symbol
        label.set_label(text)
        for c in ("dim-label", "success", "error"):
            label.remove_css_class(c)
        label.add_css_class(css)

    # ─── Context menu ───────────────────────────────────────

    def _on_list_right_click(self, gesture, n_press, x, y):
        outer = self._find_outer_at(x, y)
        if not outer:
            return
        item = outer._chat_item
        if (
            item
            and item.kind == ChatItem.KIND_MESSAGE
            and item.message
            and item.message.is_outgoing
        ):
            gesture.set_state(Gtk.EventSequenceState.CLAIMED)
            self._context_item = item
            self._context_menu_open = True
            self._is_sticky = False
            self._frozen_scroll_value = self._scroll.get_vadjustment().get_value()
            rect = Gdk.Rectangle()
            rect.x, rect.y, rect.width, rect.height = int(x), int(y), 1, 1
            self._context_popover.set_pointing_to(rect)
            self._context_popover.popup()

    def _resend_context_item(self):
        if self._context_item and self._context_item.message:
            self.resend_message(self._context_item.message)

    def _delete_context_item(self):
        if self._context_item and self._context_item.message:
            self.delete_message(self._context_item.message)

    def _on_share_contact(self, action, param):
        if not self._contact:
            return
        contacts = [
            c for c in self._window.contacts if c.public_key_hex != self._contact.public_key_hex
        ]
        contact = self._contact

        def _send(c):
            text = f"<{c.public_key_hex}:{c.type}:{c.name}>"
            self._window.send_message(contact, text)

        share_contact_dialog(self._window, contacts, _send)

    def _on_share_location(self, action, param):
        if not self._contact:
            return
        contact = self._contact
        share_location(self._window, lambda text: self._window.send_message(contact, text))

    def _on_share_location_map(self, action, param):
        if not self._contact:
            return
        contact = self._contact
        share_location_map(self._window, lambda text: self._window.send_message(contact, text))

    # ─── Contact & messages ──────────────────────────────────

    def set_contact(self, contact: Contact):
        if self._contact:
            adj = self._scroll.get_vadjustment()
            self._scroll_positions[self._contact.public_key_hex] = adj.get_value()

        self._contact = contact
        self._stack.set_visible_child_name("chat")
        self._load_messages()

    def _load_messages(self):
        self._item_by_msg_id.clear()
        self._unread_pos = -1
        self._messages_offset = 0
        self._all_loaded = False
        self._loading_more = False

        if not self._contact:
            self._store.remove_all()
            return

        messages = self._window.get_messages_for_contact(self._contact, limit=_BATCH_SIZE)
        self._messages_offset = len(messages)
        self._all_loaded = len(messages) < _BATCH_SIZE
        last_read_at = self._window.get_contact_last_read_at(self._contact.public_key_hex)

        new_items = []
        divider_inserted = False
        last_date = None
        for msg in messages:
            msg_date = msg.timestamp.date()
            if last_date is not None and msg_date != last_date:
                new_items.append(
                    ChatItem(
                        kind=ChatItem.KIND_DATE_DIVIDER,
                        date_text=_format_date_text(msg.timestamp),
                    )
                )
            last_date = msg_date

            # Insert unread divider before first unread incoming message
            if (
                not divider_inserted
                and last_read_at is not None
                and not msg.is_outgoing
                and msg.timestamp.timestamp() > last_read_at
            ):
                new_items.append(ChatItem(kind=ChatItem.KIND_UNREAD_DIVIDER))
                self._unread_pos = len(new_items) - 1
                divider_inserted = True

            item = ChatItem(kind=ChatItem.KIND_MESSAGE, message=msg)
            new_items.append(item)
            if msg.message_id:
                self._item_by_msg_id[msg.message_id] = item

        saved_pos = self._scroll_positions.get(self._contact.public_key_hex)
        if saved_pos is not None:
            self._is_sticky = False
            self._restoring_scroll = True

        # Replace all items in a single atomic operation
        self._store.splice(0, self._store.get_n_items(), new_items)

        if saved_pos is not None:
            GLib.idle_add(lambda pos=saved_pos: self._restore_scroll_position(pos))
        else:
            GLib.idle_add(self._scroll_to_bottom)
        if self._unread_pos >= 0:
            GLib.idle_add(lambda: self._update_unread_btn_visibility() or False)

        # Mark as read now
        self._window.mark_contact_read(self._contact.public_key_hex)

    def _load_older_messages(self):
        if self._all_loaded or self._loading_more or not self._contact:
            return
        self._loading_more = True

        messages = self._window.get_messages_for_contact(
            self._contact, limit=_BATCH_SIZE, offset=self._messages_offset
        )
        if not messages:
            self._all_loaded = True
            self._loading_more = False
            return

        older_items = []
        last_date = None
        for msg in messages:
            msg_date = msg.timestamp.date()
            if last_date is not None and msg_date != last_date:
                older_items.append(
                    ChatItem(
                        kind=ChatItem.KIND_DATE_DIVIDER,
                        date_text=_format_date_text(msg.timestamp),
                    )
                )
            last_date = msg_date
            item = ChatItem(kind=ChatItem.KIND_MESSAGE, message=msg)
            older_items.append(item)
            if msg.message_id:
                self._item_by_msg_id[msg.message_id] = item

        # Add boundary date divider if the dates differ
        if last_date is not None:
            for i in range(self._store.get_n_items()):
                first = self._store.get_item(i)
                if first.kind == ChatItem.KIND_MESSAGE:
                    if first.message.timestamp.date() != last_date:
                        older_items.append(
                            ChatItem(
                                kind=ChatItem.KIND_DATE_DIVIDER,
                                date_text=_format_date_text(first.message.timestamp),
                            )
                        )
                    break

        self._store.splice(0, 0, older_items)
        if self._unread_pos >= 0:
            self._unread_pos += len(older_items)
        self._messages_offset += len(messages)
        self._all_loaded = len(messages) < _BATCH_SIZE
        self._loading_more = False

    def _on_text_changed(self, entry):
        text = entry.get_text()
        used = len(text.encode("utf-8"))
        limit = max_dm_text_bytes()
        if used == 0:
            self._char_label.set_label("")
            self._send_btn.set_sensitive(True)
            self._char_label.remove_css_class("error")
        else:
            self._char_label.set_label(f"{used}/{limit}")
            over = used > limit
            self._send_btn.set_sensitive(not over)
            if over:
                self._char_label.add_css_class("error")
            else:
                self._char_label.remove_css_class("error")

    def _on_send(self, *args):
        if not self._contact:
            return
        text = self._text_entry.get_text().strip()
        if not text:
            return
        if len(text.encode("utf-8")) > max_dm_text_bytes():
            return
        self._text_entry.set_text("")
        self._window.send_message(self._contact, text)

    def _on_send_with_override(self, route_override: str | None):
        if not self._contact:
            return
        text = self._text_entry.get_text().strip()
        if not text:
            return
        if len(text.encode("utf-8")) > max_dm_text_bytes():
            return
        self._text_entry.set_text("")
        self._window.send_message(self._contact, text, route_override=route_override)

    def on_message_received(self, contact: Contact, message: Message):
        """Called when a new message is received."""
        if self._contact and contact.public_key_hex == self._contact.public_key_hex:
            item = ChatItem(kind=ChatItem.KIND_MESSAGE, message=message)
            # If sticky, the notify::upper handler will auto-scroll
            self._store.append(item)

    def on_message_sent(self, message: Message):
        """Called when a message is sent."""
        # Force sticky so notify::upper handler scrolls to the new message
        self._is_sticky = True
        item = ChatItem(kind=ChatItem.KIND_MESSAGE, message=message)
        self._store.append(item)
        if message.message_id:
            self._item_by_msg_id[message.message_id] = item

    def resend_message(self, message: Message):
        """Re-send a message to the current contact."""
        if self._contact and message.message_id:
            self._window.resend_message(self._contact, message)

    def delete_message(self, message: Message):
        """Delete a message from the view and storage."""
        if not message.message_id:
            return
        for i in range(self._store.get_n_items()):
            item = self._store.get_item(i)
            if (
                item.kind == ChatItem.KIND_MESSAGE
                and item.message
                and item.message.message_id == message.message_id
            ):
                self._store.remove(i)
                break
        self._item_by_msg_id.pop(message.message_id, None)
        self._window.delete_message(message.message_id)

    def update_message_status(self, message_id: str, status: MessageStatus, detail: str = ""):
        """Update the status of a sent message."""
        if message_id in self._item_by_msg_id:
            item = self._item_by_msg_id[message_id]
            item.status_detail = detail
            # Setting status triggers notify::status -> bound widget updates
            item.props.status = int(status)



# Copyright 2026, Jiri Eischmann <jiri@eischmann.cz> and the meshy contributors
# SPDX-License-Identifier: GPL-3.0-or-later

import math
import re
import unicodedata

from gi.repository import GLib

_URL_RE = re.compile(r'(https?://[^\s<>"\')\]]+)')
_SHARED_CONTACT_RE = re.compile(r"^<([0-9a-fA-F]{64}):(\d+):(.+)>$")
_SHARED_LOCATION_RE = re.compile(r"^(-?\d+\.\d+),\s*(-?\d+\.\d+)$")


def linkify(text: str) -> str:
    escaped = GLib.markup_escape_text(text)

    def _replace(m):
        url = m.group(1)
        trailing = ""
        while url and url[-1] in ".,;:!?":
            trailing = url[-1] + trailing
            url = url[:-1]
        return f'<a href="{url}">{url}</a>{trailing}'

    return _URL_RE.sub(_replace, escaped)


def parse_shared_contact(text: str) -> tuple[str, int, str] | None:
    m = _SHARED_CONTACT_RE.match(text.strip())
    if not m:
        return None
    return (m.group(1), int(m.group(2)), m.group(3))


def parse_shared_location(text: str) -> tuple[float, float] | None:
    m = _SHARED_LOCATION_RE.match(text.strip())
    if not m:
        return None
    lat, lon = float(m.group(1)), float(m.group(2))
    if -90 <= lat <= 90 and -180 <= lon <= 180:
        return (lat, lon)
    return None


def haversine_distance(lat1, lon1, lat2, lon2) -> float:
    """Return the great-circle distance in km between two points."""
    R = 6371.0
    dlat = math.radians(lat2 - lat1)
    dlon = math.radians(lon2 - lon1)
    a = (
        math.sin(dlat / 2) ** 2
        + math.cos(math.radians(lat1)) * math.cos(math.radians(lat2)) * math.sin(dlon / 2) ** 2
    )
    return R * 2 * math.atan2(math.sqrt(a), math.sqrt(1 - a))


def extract_emoji(name: str) -> str | None:
    if not name:
        return None
    for char in name:
        cat = unicodedata.category(char)
        if cat == "So":
            return char
        cp = ord(char)
        if 0x1F600 <= cp <= 0x1F64F:
            return char
        if 0x1F300 <= cp <= 0x1F5FF:
            return char
        if 0x1F680 <= cp <= 0x1F6FF:
            return char
        if 0x1F900 <= cp <= 0x1F9FF:
            return char
        if 0x1FA00 <= cp <= 0x1FAFF:
            return char
        if 0x2600 <= cp <= 0x27BF:
            return char
        if 0xFE00 <= cp <= 0xFE0F:
            continue
        if cp == 0x200D:
            continue
    return None



# Copyright 2026, Jiri Eischmann <jiri@eischmann.cz> and the meshy contributors
# SPDX-License-Identifier: GPL-3.0-or-later

"""Contact detail dialog — shows info, routing, telemetry, and actions for a contact."""

import gi

gi.require_version("Gtk", "4.0")
gi.require_version("Adw", "1")
from gi.repository import Adw, GLib, Gtk
from meshy.models import Contact, ContactType


@Gtk.Template(resource_path="/page/codeberg/sesivany/Meshy/ui/contact-detail-dialog.ui")
class ContactDetailDialog(Adw.Dialog):
    __gtype_name__ = "MeshyContactDetailDialog"

    content_box = Gtk.Template.Child()
    info_group = Gtk.Template.Child()
    name_entry = Gtk.Template.Child()
    favorite_switch = Gtk.Template.Child()
    type_row = Gtk.Template.Child()
    key_row = Gtk.Template.Child()
    copy_btn = Gtk.Template.Child()
    last_seen_row = Gtk.Template.Child()
    routing_group = Gtk.Template.Child()
    routing_toggle = Gtk.Template.Child()
    out_path_entry = Gtk.Template.Child()
    reset_path_row = Gtk.Template.Child()
    reset_path_btn = Gtk.Template.Child()
    smart_paths_row = Gtk.Template.Child()
    auto_reset_switch = Gtk.Template.Child()
    flood_switch = Gtk.Template.Child()
    tele_group = Gtk.Template.Child()
    req_tele_row = Gtk.Template.Child()
    req_tele_btn = Gtk.Template.Child()
    danger_group = Gtk.Template.Child()
    remove_btn = Gtk.Template.Child()

    def __init__(self, contacts_view, contact: Contact, **kwargs):
        super().__init__(**kwargs)
        self._contacts_view = contacts_view
        self._window = contacts_view._window
        self._contact = contact
        self._path_map_row = None
        self._updating = False

        self.set_title(contact.name)
        self._populate(contact)
        self._add_conditional_widgets(contact)
        self._connect_signals(contact)

    def _populate(self, contact: Contact):
        self.name_entry.set_text(contact.name)
        self.favorite_switch.set_active(contact.is_favorite)
        self.type_row.set_subtitle(contact.type_label)
        self.key_row.set_subtitle(contact.public_key_hex)
        last_seen_str = (
            contact.last_seen.strftime("%Y-%m-%d %H:%M") if contact.last_seen else _("Never")
        )
        self.last_seen_row.set_subtitle(last_seen_str)

        # Routing state
        is_smart = contact.smart_route_selection
        self.routing_toggle.set_active_name("smart" if is_smart else "classic")
        if contact.path_hops:
            self.out_path_entry.set_text(",".join(h.lower() for h in contact.path_hops))
        self.out_path_entry.set_visible(not is_smart)
        self.reset_path_row.set_visible(not is_smart)
        self.reset_path_row.set_activatable_widget(self.reset_path_btn)
        self.smart_paths_row.set_visible(is_smart)
        self.auto_reset_switch.set_active(contact.auto_reset_path)
        self.flood_switch.set_active(contact.force_flood)
        self.flood_switch.set_visible(not is_smart)
        self._update_smart_subtitle(contact)

        self.req_tele_row.set_activatable_widget(self.req_tele_btn)

    def _add_conditional_widgets(self, contact: Contact):
        # Location row in info group
        if contact.has_location:
            loc_row = Adw.ActionRow(
                title=_("Location"),
                subtitle=f"{contact.latitude:.6f}, {contact.longitude:.6f}",
            )
            map_btn = Gtk.Button(icon_name="map-symbolic", valign=Gtk.Align.CENTER)
            map_btn.add_css_class("flat")
            map_btn.connect("clicked", lambda *_, c=contact: self._contacts_view._show_on_map(c))
            loc_row.add_suffix(map_btn)
            loc_row.set_activatable_widget(map_btn)
            self.info_group.add(loc_row)

        # Show Path on Map in routing group
        if contact.path_hops:
            hop_contacts = self._window.resolve_hop_prefixes(contact.path_hops)
            has_located = any(c and c.has_location for c in hop_contacts)
            if has_located or contact.has_location:
                self._path_map_row = Adw.ActionRow(
                    title=_("Show Path on Map"),
                    subtitle=_("Visualize the message route on the map"),
                )
                path_map_btn = Gtk.Button(icon_name="map-symbolic", valign=Gtk.Align.CENTER)
                path_map_btn.add_css_class("flat")
                path_map_btn.connect(
                    "clicked",
                    lambda *_, c=contact, hc=hop_contacts, hp=contact.path_hops: (
                        self._contacts_view._show_path_on_map(c, hc, hp)
                    ),
                )
                self._path_map_row.add_suffix(path_map_btn)
                self._path_map_row.set_activatable_widget(path_map_btn)
                self._path_map_row.set_visible(not contact.smart_route_selection)
                self.routing_group.add(self._path_map_row)

        # Telemetry switches (only for CHAT contacts)
        if contact.type == ContactType.CHAT:
            from meshy.models import (
                CONTACT_FLAG_TELE_BASE,
                CONTACT_FLAG_TELE_ENV,
                CONTACT_FLAG_TELE_LOC,
            )

            for title, subtitle, flag in (
                (
                    _("Allow Telemetry Requests"),
                    _("Allow this contact to query battery, voltage, temperature"),
                    CONTACT_FLAG_TELE_BASE,
                ),
                (
                    _("Include Location"),
                    _("Include GPS coordinates in telemetry responses"),
                    CONTACT_FLAG_TELE_LOC,
                ),
                (
                    _("Include Environment Sensors"),
                    _("Include humidity, pressure, air quality data"),
                    CONTACT_FLAG_TELE_ENV,
                ),
            ):
                sw = Adw.SwitchRow(title=title, subtitle=subtitle)
                sw.set_active(bool(contact.flags & flag))

                def _on_toggle(*args, _sw=sw, _flag=flag):
                    if _sw.get_active():
                        contact.flags |= _flag
                    else:
                        contact.flags &= ~_flag
                    self._window.storage.save_contact(contact)
                    self._window.update_contact_flags(contact)

                sw.connect("notify::active", _on_toggle)
                self.tele_group.add(sw)

        # Actions group (only for repeaters/rooms)
        actions_count = 0
        actions_group = Adw.PreferencesGroup(title=_("Actions"))

        if contact.type == ContactType.ROOM:
            mgmt_row = Adw.ActionRow(
                title=_("Room Management"),
                subtitle=_("Login and access CLI, status, and settings"),
            )
            mgmt_btn = Gtk.Button(icon_name="utilities-terminal-symbolic", valign=Gtk.Align.CENTER)
            mgmt_btn.connect(
                "clicked",
                lambda *_: (
                    self.close(),
                    self._contacts_view._show_room_login(contact, management=True),
                ),
            )
            mgmt_row.add_suffix(mgmt_btn)
            mgmt_row.set_activatable_widget(mgmt_btn)
            actions_group.add(mgmt_row)
            actions_count += 1

        if contact.type == ContactType.REPEATER:
            mgmt_row = Adw.ActionRow(
                title=_("Repeater Management"),
                subtitle=_("Login and access status, CLI, neighbors, and settings"),
            )
            mgmt_btn = Gtk.Button(icon_name="utilities-terminal-symbolic", valign=Gtk.Align.CENTER)
            mgmt_btn.connect(
                "clicked",
                lambda *_: (
                    self.close(),
                    self._contacts_view._show_room_login(contact, repeater=True),
                ),
            )
            mgmt_row.add_suffix(mgmt_btn)
            mgmt_row.set_activatable_widget(mgmt_btn)
            actions_group.add(mgmt_row)
            actions_count += 1

        if contact.type in (ContactType.REPEATER, ContactType.ROOM):
            # Ping
            ping_row = Adw.ActionRow(
                title=_("Ping"),
                subtitle=_("Measure round-trip time to this node"),
            )
            ping_btn = Gtk.Button(
                icon_name="network-transmit-receive-symbolic", valign=Gtk.Align.CENTER
            )
            ping_btn.connect(
                "clicked", lambda *_: self._contacts_view._ping_contact(contact, ping_btn)
            )
            ping_row.add_suffix(ping_btn)
            ping_row.set_activatable_widget(ping_btn)
            actions_group.add(ping_row)
            actions_count += 1

            # Discover Paths
            discover_row = Adw.ActionRow(
                title=_("Discover Paths"),
                subtitle=_("Find routes to this node via flood"),
            )
            self._contacts_view._discover_path_label = Gtk.Label(label="")
            self._contacts_view._discover_path_label.add_css_class("dim-label")
            self._contacts_view._discover_path_label.set_valign(Gtk.Align.CENTER)
            discover_row.add_suffix(self._contacts_view._discover_path_label)
            discover_btn = Gtk.Button(icon_name="system-search-symbolic", valign=Gtk.Align.CENTER)
            discover_btn.connect(
                "clicked", lambda *_: self._contacts_view._on_path_discovery(contact, discover_btn)
            )
            discover_row.add_suffix(discover_btn)
            discover_row.set_activatable_widget(discover_btn)
            actions_group.add(discover_row)
            actions_count += 1

            # Regions
            regions_row = Adw.ActionRow(
                title=_("Regions"),
                subtitle=_("Query regions from this repeater"),
            )
            regions_label = Gtk.Label(label="")
            regions_label.add_css_class("dim-label")
            regions_label.set_valign(Gtk.Align.CENTER)
            regions_row.add_suffix(regions_label)
            regions_btn = Gtk.Button(icon_name="region-symbolic", valign=Gtk.Align.CENTER)

            def _on_regions_click(*_args, _contact=contact, _btn=regions_btn, _label=regions_label):
                _btn.set_sensitive(False)
                _label.set_label(_("Querying..."))

                def _on_response(text):
                    def _update():
                        _btn.set_sensitive(True)
                        if text:
                            names = []
                            for line in text.replace(",", "\n").split("\n"):
                                name = line.strip().strip("#")
                                if name and name != "*":
                                    names.append(name)
                            _label.set_label(", ".join(names) if names else _("None"))
                        else:
                            _label.set_label(_("No response"))

                    GLib.idle_add(_update)

                responded = [False]

                def _handle(text):
                    if responded[0]:
                        return
                    responded[0] = True
                    _on_response(text)

                self._window.request_regions(_contact, lambda text: _handle(text), use_path=True)

                def _retry():
                    if responded[0]:
                        return False
                    self._window.request_regions(
                        _contact, lambda text: _handle(text), use_path=True
                    )
                    return False

                def _give_up():
                    _handle(None)
                    return False

                GLib.timeout_add(5000, _retry)
                GLib.timeout_add(10000, _give_up)

            regions_btn.connect("clicked", _on_regions_click)
            regions_row.add_suffix(regions_btn)
            regions_row.set_activatable_widget(regions_btn)
            actions_group.add(regions_row)
            actions_count += 1

        if actions_count > 0:
            self.content_box.insert_child_after(actions_group, self.tele_group)

    def _connect_signals(self, contact: Contact):
        # Info signals
        self.favorite_switch.connect("notify::active", self._on_favorite_toggled, contact)
        self.copy_btn.connect("clicked", self._on_copy_key, contact)
        self.name_entry.connect("apply", self._on_apply, contact)
        self.out_path_entry.connect("apply", self._on_apply, contact)

        # Routing signals
        self.routing_toggle.connect("notify::active-name", self._on_toggle_changed, contact)
        self.out_path_entry.connect("changed", self._on_path_changed)
        self.reset_path_btn.connect("clicked", self._on_reset_path, contact)
        self.smart_paths_row.connect("activated", self._on_smart_paths_activated, contact)
        self.flood_switch.connect("notify::active", self._on_flood_toggled, contact)
        self.flood_switch.connect(
            "notify::active",
            lambda *a: self._on_apply(None, contact) if not self._updating else None,
        )
        self.auto_reset_switch.connect("notify::active", self._on_auto_reset_toggled, contact)

        # Telemetry
        self.req_tele_btn.connect(
            "clicked",
            lambda *_: self._contacts_view._request_telemetry(contact, self.req_tele_row),
        )

        # Danger zone
        self.remove_btn.connect("clicked", self._on_remove, contact)

        # Initial state
        self._update_flood_dependents()

    # --- Info callbacks ---

    def _on_favorite_toggled(self, switch, _pspec, contact):
        from meshy.models import CONTACT_FLAG_FAVORITE

        if switch.get_active():
            contact.flags |= CONTACT_FLAG_FAVORITE
        else:
            contact.flags &= ~CONTACT_FLAG_FAVORITE
        self._window.storage.save_contact(contact)
        self._window.update_contact_flags(contact)
        self._window.update_contacts_from_storage()

    def _on_copy_key(self, _btn, contact):
        clipboard = self._window.get_clipboard()
        clipboard.set(contact.public_key_hex)
        clipboard.store_async(None, None)
        self._window.show_toast(_("Public key copied"))

    # --- Routing callbacks ---

    def _on_toggle_changed(self, toggle_group, _pspec, contact):
        if self._updating:
            return
        is_smart = toggle_group.get_active_name() == "smart"
        contact.smart_route_selection = is_smart
        self._window.storage.save_contact(contact)
        self.out_path_entry.set_visible(not is_smart)
        self.reset_path_row.set_visible(not is_smart)
        self.flood_switch.set_visible(not is_smart)
        if self._path_map_row is not None:
            self._path_map_row.set_visible(not is_smart)
        self.smart_paths_row.set_visible(is_smart)
        if is_smart:
            self._update_smart_subtitle(contact)
            self._update_flood_dependents()

    def _on_path_changed(self, *args):
        if self._updating:
            return
        if self.out_path_entry.get_text().strip():
            self._updating = True
            self.flood_switch.set_active(False)
            self._updating = False

    def _on_reset_path(self, _btn, contact):
        self._window.reset_contact_path(contact)
        self.out_path_entry.set_text("")

    def _on_smart_paths_activated(self, _row, contact):
        self._show_smart_paths_dialog(contact)

    def _on_flood_toggled(self, switch, _pspec, contact):
        if self._updating:
            return
        self._update_flood_dependents()
        if switch.get_active():
            self._updating = True
            self.out_path_entry.set_text("")
            self._updating = False
        contact.force_flood = switch.get_active()
        self._window.storage.save_contact(contact)

    def _on_auto_reset_toggled(self, switch, _pspec, contact):
        if self._updating:
            return
        contact.auto_reset_path = switch.get_active()
        self._window.storage.save_contact(contact)

    def _update_flood_dependents(self):
        is_flood = self.flood_switch.get_active()
        self.auto_reset_switch.set_sensitive(not is_flood)

    def _update_smart_subtitle(self, contact):
        count = len(self._window.storage.get_path_history(contact.public_key_hex))
        self.smart_paths_row.set_subtitle(
            _("%d paths available") % count if count else _("No paths yet")
        )

    # --- Apply ---

    def _on_apply(self, _widget, contact):
        new_name = self.name_entry.get_text().strip() or contact.name
        path_text = self.out_path_entry.get_text().strip()
        force_flood = self.flood_switch.get_active()

        if force_flood:
            self._window.set_contact_path(contact, 0xFF, b"", name=new_name)
        elif path_text:
            try:
                hops = [h.strip() for h in path_text.split(",") if h.strip()]
                hash_size = len(bytes.fromhex(hops[0]))
                path_bytes = b"".join(bytes.fromhex(h) for h in hops)
                hash_mode = hash_size - 1
                hop_count = len(hops)
                path_len_byte = (hash_mode << 6) | (hop_count & 0x3F)
                self._window.set_contact_path(contact, path_len_byte, path_bytes, name=new_name)
            except ValueError:
                pass
        else:
            self._window.reset_contact_path(contact)
            if new_name != contact.name:
                GLib.timeout_add(
                    500,
                    lambda: (
                        self._window.set_contact_path(contact, 0xFF, b"", name=new_name) or False
                    ),
                )

        if new_name != contact.name:
            contact.name = new_name
            self._window.storage.save_contact(contact)
            self._window.update_contacts_from_storage()
            self._window.refresh_chat_header()

    # --- Smart paths dialog ---

    def _show_smart_paths_dialog(self, contact: Contact):
        from meshy.views.map_view import _ROUTE_COLORS

        path_svc = self._window._msg_ctrl.path_history
        records = self._window.storage.get_path_history(contact.public_key_hex)

        dialog = Adw.Dialog()
        dialog.set_title(_("Smart Routes — %s") % contact.name)
        dialog.set_content_width(420)
        dialog.set_content_height(360)

        toolbar_view = Adw.ToolbarView()
        header = Adw.HeaderBar()
        if records:
            map_btn = Gtk.Button(icon_name="map-symbolic", tooltip_text=_("Show on Map"))
            map_btn.add_css_class("flat")
            map_btn.connect(
                "clicked",
                lambda *_, d=dialog: self._show_all_routes_on_map(contact, records, path_svc, d),
            )
            header.pack_end(map_btn)
        toolbar_view.add_top_bar(header)

        scroll = Gtk.ScrolledWindow(vexpand=True)
        clamp = Adw.Clamp(maximum_size=600)

        box = Gtk.Box(
            orientation=Gtk.Orientation.VERTICAL,
            spacing=12,
            margin_start=16,
            margin_end=16,
            margin_top=12,
            margin_bottom=12,
        )

        paths_group = Adw.PreferencesGroup(title=_("Available Paths"))

        if records:
            for i, r in enumerate(sorted(records, key=path_svc.score_dict, reverse=True)):
                path_hex = r["path_bytes"]
                if path_hex:
                    chunk = r["path_hash_size"] * 2
                    hops = ",".join(path_hex[j : j + chunk] for j in range(0, len(path_hex), chunk))
                else:
                    hops = _("Direct")
                sub_parts = [
                    _("Score: %.2f") % path_svc.score_dict(r),
                    _("Weight: %.1f") % r["route_weight"],
                    f"{r['success_count']}✓ {r['failure_count']}✗",
                    _("%d hops") % r["hop_count"],
                ]
                if r.get("last_trip_ms"):
                    sub_parts.append(_("%d ms") % r["last_trip_ms"])
                row = Adw.ActionRow(title=hops, subtitle=" · ".join(sub_parts))
                cr, cg, cb = _ROUTE_COLORS[i % len(_ROUTE_COLORS)]
                dot = Gtk.DrawingArea()
                dot.set_content_width(12)
                dot.set_content_height(12)
                dot.set_valign(Gtk.Align.CENTER)
                dot.set_draw_func(
                    lambda _area, ctx, w, h, _r=cr, _g=cg, _b=cb: (
                        ctx.set_source_rgb(_r, _g, _b),
                        ctx.arc(w / 2, h / 2, min(w, h) / 2, 0, 2 * 3.141593),
                        ctx.fill(),
                    )
                )
                row.add_prefix(dot)
                paths_group.add(row)
        else:
            empty_row = Adw.ActionRow(
                title=_("No paths recorded yet"),
                subtitle=_("Paths are discovered as messages are exchanged"),
            )
            paths_group.add(empty_row)

        box.append(paths_group)

        if records:
            clear_btn = Gtk.Button(label=_("Clear History"))
            clear_btn.add_css_class("destructive-action")
            clear_btn.set_halign(Gtk.Align.CENTER)

            def _on_clear(*_args):
                self._window.storage.clear_path_history(contact.public_key_hex)
                self.smart_paths_row.set_subtitle(_("No paths yet"))
                dialog.close()

            clear_btn.connect("clicked", _on_clear)
            box.append(clear_btn)

        clamp.set_child(box)
        scroll.set_child(clamp)
        toolbar_view.set_content(scroll)
        dialog.set_child(toolbar_view)
        dialog.present(self._window)

    # --- Smart routes map ---

    def _show_all_routes_on_map(self, contact, records, path_svc, parent_dialog):
        from gi.repository import Shumate
        from meshy.views.map_view import (
            _MARKER_COLORS,
            _MARKER_ICONS,
            _ROUTE_COLORS,
            _SELF_COLOR,
            _UNKNOWN_COLOR,
            MapView,
            companion_too_far,
        )

        sorted_records = sorted(records, key=path_svc.score_dict, reverse=True)

        device = self._window.device_info
        self_lat = getattr(device, "latitude", None)
        self_lon = getattr(device, "longitude", None)
        has_self = self_lat and self_lon and (abs(self_lat) > 1e-6 or abs(self_lon) > 1e-6)

        route_hops = []
        for i, rec in enumerate(sorted_records):
            color_rgb = _ROUTE_COLORS[i % len(_ROUTE_COLORS)]
            path_hex = rec["path_bytes"]

            raw_hops = []

            if path_hex:
                chunk = rec["path_hash_size"] * 2
                hop_hexes = [path_hex[j : j + chunk] for j in range(0, len(path_hex), chunk)]
                hop_contacts = self._window.resolve_hop_prefixes(hop_hexes)
                for j, hc in enumerate(hop_contacts):
                    prefix = hop_hexes[j] if j < len(hop_hexes) else None
                    if hc and hc.has_location:
                        raw_hops.append((hc.latitude, hc.longitude, hc, prefix, None))
                    else:
                        raw_hops.append((None, None, hc, prefix, None))

            if contact.has_location:
                raw_hops.append((contact.latitude, contact.longitude, contact, None, None))

            if has_self and not companion_too_far(self_lat, self_lon, raw_hops):
                raw_hops.insert(0, (self_lat, self_lon, None, None, None))

            hops = MapView.interpolate_unknown_hops(raw_hops)
            route_hops.append((color_rgb, hops))

        all_lats, all_lons = [], []
        for _color, hops in route_hops:
            for h in hops:
                if h[0] is not None:
                    all_lats.append(h[0])
                    all_lons.append(h[1])

        if not all_lats:
            self._window.show_toast(_("No location data available"))
            return

        dialog, map_widget, marker_layer, viewport = MapView.open_map_dialog(
            self._window, title=_("Routes to %s") % contact.name
        )

        map_widget.get_map().remove_layer(marker_layer)

        for color_rgb, hops in route_hops:
            path_layer = MapView.create_dashed_path_layer(
                viewport, color_rgb=color_rgb, stroke_width=3
            )
            for lat, lon, _hc, _pfx, _snr2 in hops:
                if lat is not None:
                    node = Shumate.Marker()
                    node.set_location(lat, lon)
                    path_layer.add_node(node)
            map_widget.get_map().add_layer(path_layer)

        map_widget.get_map().add_layer(marker_layer)

        placed = set()
        for _color, hops in route_hops:
            for lat, lon, hop_contact, prefix, _snr in hops:
                if lat is None:
                    continue
                key = (round(lat, 6), round(lon, 6))
                if key in placed:
                    continue
                placed.add(key)

                if hop_contact is None and prefix is None:
                    r, g, b = _SELF_COLOR
                    icon_name = "network-wireless-symbolic"
                    name = device.name or _("My Device")
                elif hop_contact is None:
                    r, g, b = _UNKNOWN_COLOR
                    icon_name = "network-server-symbolic"
                    name = prefix or _("Unknown")
                else:
                    r, g, b = _MARKER_COLORS.get(hop_contact.type, (0.5, 0.5, 0.5))
                    icon_name = _MARKER_ICONS.get(hop_contact.type, "avatar-default-symbolic")
                    name = hop_contact.name

                marker = MapView.create_marker(r, g, b, icon_name, name, lat, lon, show_label=True)
                marker_layer.add_marker(marker)

        # Legend bar
        toolbar_view = dialog.get_child()
        legend_scroll = Gtk.ScrolledWindow(
            hscrollbar_policy=Gtk.PolicyType.AUTOMATIC,
            vscrollbar_policy=Gtk.PolicyType.NEVER,
        )
        legend_box = Gtk.Box(
            spacing=16,
            margin_start=8,
            margin_end=8,
            margin_top=4,
            margin_bottom=4,
        )
        for i, rec in enumerate(sorted_records):
            cr, cg, cb = _ROUTE_COLORS[i % len(_ROUTE_COLORS)]
            item = Gtk.Box(spacing=4)
            dot = Gtk.DrawingArea()
            dot.set_content_width(10)
            dot.set_content_height(10)
            dot.set_valign(Gtk.Align.CENTER)
            dot.set_draw_func(
                lambda _area, ctx, w, h, _r=cr, _g=cg, _b=cb: (
                    ctx.set_source_rgb(_r, _g, _b),
                    ctx.arc(w / 2, h / 2, min(w, h) / 2, 0, 2 * 3.141593),
                    ctx.fill(),
                )
            )
            item.append(dot)
            path_hex = rec["path_bytes"]
            if path_hex:
                label_text = _("%d hops") % rec["hop_count"]
            else:
                label_text = _("Direct")
            lbl = Gtk.Label(label=label_text)
            lbl.add_css_class("caption")
            item.append(lbl)
            legend_box.append(item)
        legend_scroll.set_child(legend_box)
        toolbar_view.add_bottom_bar(legend_scroll)

        MapView.fit_viewport(viewport, all_lats, all_lons)
        dialog.present(parent_dialog)

    # --- Danger zone ---

    def _on_remove(self, _btn, contact):
        confirm = Adw.AlertDialog(
            heading=_("Remove Contact?"),
            body=_("Remove {} and all messages?").format(contact.name),
        )
        confirm.add_response("cancel", _("Cancel"))
        confirm.add_response("remove", _("Remove"))
        confirm.set_response_appearance("remove", Adw.ResponseAppearance.DESTRUCTIVE)
        confirm.connect(
            "response",
            lambda d, r: (
                (self._window.remove_contact(contact), self.close()) if r == "remove" else None
            ),
        )
        confirm.present(self._window)



import math

from meshy.utils import (
    extract_emoji,
    haversine_distance,
    parse_shared_contact,
    parse_shared_location,
)

# ─── parse_shared_contact ─────────────────────────────────────


class TestParseSharedContact:
    def test_valid(self):
        key = "aa" * 32
        result = parse_shared_contact(f"<{key}:1:Alice>")
        assert result == (key, 1, "Alice")

    def test_with_spaces_in_name(self):
        key = "bb" * 32
        result = parse_shared_contact(f"<{key}:2:Bob Smith>")
        assert result == (key, 2, "Bob Smith")

    def test_invalid_format(self):
        assert parse_shared_contact("not a contact") is None

    def test_short_key(self):
        assert parse_shared_contact("<aabb:1:Test>") is None


# ─── parse_shared_location ────────────────────────────────────


class TestParseSharedLocation:
    def test_valid(self):
        result = parse_shared_location("50.123456, 14.456789")
        assert result is not None
        assert abs(result[0] - 50.123456) < 1e-6
        assert abs(result[1] - 14.456789) < 1e-6

    def test_negative_coords(self):
        result = parse_shared_location("-33.868820, 151.209290")
        assert result is not None

    def test_out_of_range_lat(self):
        assert parse_shared_location("100.0, 14.0") is None

    def test_out_of_range_lon(self):
        assert parse_shared_location("50.0, 200.0") is None

    def test_invalid_format(self):
        assert parse_shared_location("not coords") is None


# ─── haversine_distance ───────────────────────────────────────


class TestHaversineDistance:
    def test_same_point(self):
        assert haversine_distance(50.0, 14.0, 50.0, 14.0) == 0.0

    def test_prague_brno(self):
        d = haversine_distance(50.0755, 14.4378, 49.1951, 16.6068)
        assert 180 < d < 200  # ~186 km

    def test_antipodes(self):
        d = haversine_distance(0.0, 0.0, 0.0, 180.0)
        assert abs(d - math.pi * 6371.0) < 1.0


# ─── extract_emoji ────────────────────────────────────────────


class TestExtractEmoji:
    def test_no_emoji(self):
        assert extract_emoji("Hello") is None

    def test_empty(self):
        assert extract_emoji("") is None

    def test_smiley(self):
        assert extract_emoji("Node 😀 test") == "😀"

    def test_weather_symbol(self):
        assert extract_emoji("☀ Sunny") == "☀"

    def test_rocket(self):
        assert extract_emoji("🚀 Launch") == "🚀"



# Copyright 2026, Jiri Eischmann <jiri@eischmann.cz> and the meshy contributors
# SPDX-License-Identifier: GPL-3.0-or-later

"""Connection screen - shown when no device is connected."""

import os

import gi

gi.require_version("Gtk", "4.0")
gi.require_version("Adw", "1")
gi.require_version("Gdk", "4.0")
from gi.repository import Adw, Gdk, Gio, GLib, Gtk
from meshy.ble import BleDevice
from meshy.host_integration import is_flatpak, run_host_command
from meshy.platforms import is_linux
from meshy.storage import Storage


@Gtk.Template(resource_path="/page/codeberg/sesivany/Meshy/ui/connection-view.ui")
class ConnectionView(Gtk.Box):
    """Connection screen with Bluetooth and USB connection options."""

    __gtype_name__ = "MeshyConnectionView"

    _paired_list = Gtk.Template.Child("paired_list")
    _pair_btn = Gtk.Template.Child("pair_btn")
    _usb_list = Gtk.Template.Child("usb_list")
    _tcp_list = Gtk.Template.Child("tcp_list")
    _tcp_add_btn = Gtk.Template.Child("tcp_add_btn")

    def __init__(self, window, **kwargs):
        super().__init__(**kwargs)
        self._window = window
        self._bt_poll_id = None

        self._pair_btn.connect("clicked", self._on_pair_new)
        self._tcp_add_btn.connect("clicked", self._on_tcp_add_new)

    def _add_remove_context_menu(self, row, on_remove):
        """Add a right-click / long-press context menu with Remove option."""
        menu_model = Gio.Menu()
        menu_model.append(_("Remove"), "row.remove")

        popover = Gtk.PopoverMenu(menu_model=menu_model)
        popover.set_parent(row)
        popover.set_has_arrow(False)
        row._context_popover = popover

        remove_action = Gio.SimpleAction.new("remove", None)
        remove_action.connect("activate", lambda *_: on_remove())
        action_group = Gio.SimpleActionGroup()
        action_group.add_action(remove_action)
        row.insert_action_group("row", action_group)

        def _show_menu(x, y):
            rect = Gdk.Rectangle()
            rect.x, rect.y, rect.width, rect.height = int(x), int(y), 1, 1
            popover.set_pointing_to(rect)
            popover.popup()

        right_click = Gtk.GestureClick(button=3)
        right_click.connect("pressed", lambda g, n, x, y: _show_menu(x, y))
        row.add_controller(right_click)

        long_press = Gtk.GestureLongPress()
        long_press.connect("pressed", lambda g, x, y: _show_menu(x, y))
        row.add_controller(long_press)

    def refresh_paired_devices(self):
        """Refresh the list of paired BLE devices, USB serial devices, and TCP companions."""
        self._refresh_tcp_companions()
        # Refresh BLE paired devices
        self._clear_list(self._paired_list)

        # Check if BLE transport is available
        if not self._window.ble_transport:
            placeholder = Adw.ActionRow(
                title=_("Bluetooth not available"),
                subtitle=_("BLE requires running from app bundle on macOS"),
                sensitive=False,
            )
            placeholder.add_prefix(Gtk.Image.new_from_icon_name("bluetooth-disabled-symbolic"))
            self._paired_list.append(placeholder)
            self._pair_btn.set_sensitive(False)
            return

        bt_enabled = self._window.ble_transport.is_bluetooth_enabled()
        self._pair_btn.set_sensitive(bt_enabled)

        if not bt_enabled:
            placeholder = Adw.ActionRow(
                title=_("Bluetooth is disabled"),
                subtitle=_("Enable Bluetooth in system settings to connect"),
                sensitive=False,
            )
            placeholder.add_prefix(Gtk.Image.new_from_icon_name("bluetooth-disabled-symbolic"))
            self._paired_list.append(placeholder)
        else:
            devices = self._window.ble_transport.get_paired_devices()

            if not devices:
                placeholder = Adw.ActionRow(
                    title=_("No paired devices"),
                    subtitle=_("Use the button below to pair a new companion"),
                    sensitive=False,
                )
                placeholder.add_prefix(Gtk.Image.new_from_icon_name("bluetooth-disabled-symbolic"))
                self._paired_list.append(placeholder)
            else:
                for device in devices:
                    row = Adw.ActionRow(
                        title=device.name,
                        subtitle=device.address,
                    )
                    row.add_prefix(Gtk.Image.new_from_icon_name("bluetooth-active-symbolic"))
                    connect_btn = Gtk.Button(
                        label=_("Connect"),
                        valign=Gtk.Align.CENTER,
                    )
                    connect_btn.add_css_class("suggested-action")
                    connect_btn.connect(
                        "clicked",
                        lambda b, d=device: self._window.connect_to_device(d),
                    )
                    row.add_suffix(connect_btn)
                    self._add_remove_context_menu(
                        row,
                        lambda d=device: self._remove_ble_device(d),
                    )
                    self._paired_list.append(row)

        self._bt_was_enabled = bt_enabled
        self._start_bt_polling()

        # Refresh USB serial devices
        child = self._usb_list.get_first_child()
        while child:
            next_child = child.get_next_sibling()
            self._usb_list.remove(child)
            child = next_child

        from meshy.usb_serial import UsbSerialTransport

        usb_transport = UsbSerialTransport()
        usb_devices = usb_transport.get_serial_devices()

        if not usb_devices:
            placeholder = Adw.ActionRow(
                title=_("No USB devices detected"),
                subtitle=_("Connect a MeshCore companion via USB"),
                sensitive=False,
            )
            placeholder.add_prefix(Gtk.Image.new_from_icon_name("drive-removable-media-symbolic"))
            self._usb_list.append(placeholder)
        else:
            for device in usb_devices:
                row = Adw.ActionRow(
                    title=device.name,
                    subtitle=device.path,
                )
                row.add_prefix(Gtk.Image.new_from_icon_name("drive-removable-media-symbolic"))
                connect_btn = Gtk.Button(
                    label=_("Connect"),
                    valign=Gtk.Align.CENTER,
                )
                connect_btn.add_css_class("suggested-action")
                connect_btn.connect(
                    "clicked",
                    lambda b, d=device, t=usb_transport: self._connect_usb(d, t),
                )
                row.add_suffix(connect_btn)
                self._usb_list.append(row)

    def _remove_ble_device(self, device):
        """Remove (unpair) a Bluetooth device with confirmation."""
        dialog = Adw.AlertDialog(
            heading=_("Remove Companion"),
            body=_("Remove {} ({})? You will need to pair again to reconnect.").format(
                device.name, device.address
            ),
        )
        dialog.add_response("cancel", _("Cancel"))
        dialog.add_response("remove", _("Remove"))
        dialog.set_response_appearance("remove", Adw.ResponseAppearance.DESTRUCTIVE)

        delete_check = Gtk.CheckButton(label=_("Delete stored data"))
        dialog.set_extra_child(delete_check)

        def _on_response(d, response):
            if response == "remove":
                if delete_check.get_active():
                    Storage.delete_device_db(device.address)
                self._window.ble_transport.remove_device(device.path)
                self.refresh_paired_devices()

        dialog.connect("response", _on_response)
        dialog.present(self._window)

    def _connect_usb(self, device, transport):
        """Connect to a USB device with error handling."""
        self._window.connect_to_device(device, transport=transport)
        # Check if connection failed with an error
        if hasattr(transport, "_connection_error") and transport._connection_error:
            error_msg = transport._connection_error
            is_perm = getattr(transport, "_is_permission_error", False)
            transport._connection_error = None
            transport._is_permission_error = False
            if is_perm:
                self._show_permission_dialog(device, transport)
            else:
                dialog = Adw.AlertDialog(
                    heading=_("USB Connection Failed"),
                    body=error_msg,
                )
                dialog.add_response("ok", _("OK"))
                dialog.present(self._window)

    def _can_spawn_host(self):
        """Check if we can use flatpak-spawn --host (not in sandbox, or have permission)."""
        try:
            kf = GLib.KeyFile()
            kf.load_from_file("/.flatpak-info", GLib.KeyFileFlags.NONE)
            try:
                talk_names = kf.get_string_list("Session Bus Policy", "org.freedesktop.Flatpak")
                return "talk" in talk_names or "own" in talk_names
            except GLib.Error:
                return False
        except GLib.Error:
            return True

    def _show_permission_dialog(self, device, transport):
        """Show permission error with option to install udev rule (Linux only)."""
        if not is_linux():
            # macOS doesn't need udev rules - /dev/cu.* devices are accessible by default
            return

        if self._can_spawn_host():
            self._show_install_rule_dialog(device, transport)
        else:
            self._show_manual_rule_dialog(device)

    def _show_install_rule_dialog(self, device, transport):
        """Show dialog with automatic udev rule installation."""
        dialog = Adw.AlertDialog(
            heading=_("USB Permission Denied"),
            body=_(
                "Cannot access {}.\n\n"
                "A udev rule is needed to grant access to USB serial "
                'devices. Click "Install Rule" to install it '
                "(requires administrator password)."
            ).format(device.path),
        )
        dialog.add_response("cancel", _("Cancel"))
        dialog.add_response("install", _("Install Rule"))
        dialog.set_response_appearance("install", Adw.ResponseAppearance.SUGGESTED)

        def _on_response(d, response):
            if response == "install":
                self._install_udev_rule(device, transport)

        dialog.connect("response", _on_response)
        dialog.present(self._window)

    def _show_manual_rule_dialog(self, device):
        """Show dialog with manual udev rule installation instructions."""
        command = (
            'echo \'SUBSYSTEM=="tty", KERNEL=="ttyACM[0-9]*", MODE="0666"\n'
            'SUBSYSTEM=="tty", KERNEL=="ttyUSB[0-9]*", MODE="0666"\''
            " | sudo tee /etc/udev/rules.d/60-meshy-serial.rules"
            " && sudo udevadm control --reload-rules"
            " && sudo udevadm trigger"
        )
        dialog = Adw.AlertDialog(
            heading=_("USB Permission Denied"),
            body=_(
                "Cannot access {}.\n\n"
                "A udev rule is needed to grant access to USB serial "
                "devices. Run the following command in a terminal:\n\n"
                "{}"
            ).format(device.path, command),
        )
        dialog.add_response("close", _("Close"))
        dialog.add_response("copy", _("Copy Command"))
        dialog.set_response_appearance("copy", Adw.ResponseAppearance.SUGGESTED)

        def _on_response(d, response):
            if response == "copy":
                clipboard = self._window.get_clipboard()
                clipboard.set(command)
                clipboard.store_async(None, None)
                self._window.show_toast(_("Command copied to clipboard"))

        dialog.connect("response", _on_response)
        dialog.present(self._window)

    def _install_udev_rule(self, device, transport):
        """Install udev rule via pkexec and retry connection."""
        import tempfile

        rule_content = (
            "# Meshy - MeshCore companion USB serial access\n"
            'SUBSYSTEM=="tty", KERNEL=="ttyACM[0-9]*", MODE="0666"\n'
            'SUBSYSTEM=="tty", KERNEL=="ttyUSB[0-9]*", MODE="0666"\n'
        )

        tmp_path = None
        try:
            # In Flatpak, /tmp is sandbox-private and invisible to host
            # commands run via flatpak-spawn. Use the cache dir which is
            # shared at the same absolute path.
            tmp_dir = GLib.get_user_cache_dir() if is_flatpak() else None
            with tempfile.NamedTemporaryFile(
                mode="w",
                suffix=".rules",
                delete=False,
                dir=tmp_dir,
            ) as tmp:
                tmp.write(rule_content)
                tmp_path = tmp.name

            dest = "/etc/udev/rules.d/60-meshy-serial.rules"
            result = run_host_command(
                [
                    "pkexec",
                    "bash",
                    "-c",
                    f'cp "$1" {dest}' " && udevadm control --reload-rules" " && udevadm trigger",
                    "--",
                    tmp_path,
                ],
                capture_output=True,
                text=True,
                timeout=60,
            )
            if result.returncode == 0:
                # Wait a moment then retry connection
                def _retry():
                    self._window.connect_to_device(device, transport=transport)
                    return False

                GLib.timeout_add(2000, _retry)
            else:
                dialog = Adw.AlertDialog(
                    heading=_("Installation Failed"),
                    body=_("Could not install udev rule:\n{}").format(result.stderr.strip()),
                )
                dialog.add_response("ok", _("OK"))
                dialog.present(self._window)
        except Exception as e:
            dialog = Adw.AlertDialog(
                heading=_("Installation Failed"),
                body=str(e),
            )
            dialog.add_response("ok", _("OK"))
            dialog.present(self._window)
        finally:
            if tmp_path and os.path.exists(tmp_path):
                os.unlink(tmp_path)

    def _clear_list(self, list_widget):
        child = list_widget.get_first_child()
        while child:
            next_child = child.get_next_sibling()
            popover = getattr(child, "_context_popover", None)
            if popover:
                popover.unparent()
            list_widget.remove(child)
            child = next_child

    def _refresh_tcp_companions(self):
        """Refresh the list of saved TCP companions."""
        self._clear_list(self._tcp_list)

        from gi.repository import Gio

        settings = Gio.Settings.new("page.codeberg.sesivany.Meshy")
        companions = settings.get_strv("tcp-companions")

        if not companions:
            placeholder = Adw.ActionRow(
                title=_("No saved companions"),
                subtitle=_("Use the button below to add one"),
                sensitive=False,
            )
            placeholder.add_prefix(Gtk.Image.new_from_icon_name("network-wireless-symbolic"))
            self._tcp_list.append(placeholder)
        else:
            for entry in companions:
                if "|" not in entry:
                    continue
                name, address = entry.split("|", 1)
                row = Adw.ActionRow(title=name, subtitle=address)
                row.add_prefix(Gtk.Image.new_from_icon_name("network-wireless-symbolic"))
                connect_btn = Gtk.Button(
                    label=_("Connect"),
                    valign=Gtk.Align.CENTER,
                )
                connect_btn.add_css_class("suggested-action")
                connect_btn.connect(
                    "clicked",
                    lambda b, a=address: self._connect_tcp_address(a),
                )
                row.add_suffix(connect_btn)
                self._add_remove_context_menu(
                    row,
                    lambda a=address: self._remove_tcp_companion(a),
                )
                self._tcp_list.append(row)

    def _connect_tcp_address(self, address: str):
        """Connect to a saved TCP companion by address (host:port)."""
        from meshy.tcp import TcpDevice, TcpTransport

        host, port_str = address.rsplit(":", 1)
        transport = TcpTransport()
        device = TcpDevice(host, int(port_str))
        self._window.connect_to_device(device, transport=transport)

    def _remove_tcp_companion(self, address: str):
        """Remove a saved TCP companion with confirmation."""
        dialog = Adw.AlertDialog(
            heading=_("Remove Companion"),
            body=_("Remove TCP companion {}?").format(address),
        )
        dialog.add_response("cancel", _("Cancel"))
        dialog.add_response("remove", _("Remove"))
        dialog.set_response_appearance("remove", Adw.ResponseAppearance.DESTRUCTIVE)

        delete_check = Gtk.CheckButton(label=_("Delete stored data"))
        dialog.set_extra_child(delete_check)

        def _on_response(d, response):
            if response == "remove":
                if delete_check.get_active():
                    Storage.delete_device_db(address)
                self._window.remove_tcp_companion(address)
                self._refresh_tcp_companions()

        dialog.connect("response", _on_response)
        dialog.present(self._window)

    def _on_tcp_add_new(self, *args):
        """Show dialog to add a new TCP companion."""
        dialog = Adw.AlertDialog(
            heading=_("Add TCP Companion"),
            body=_("Enter the IP address and port of your MeshCore device."),
        )

        content = Gtk.Box(orientation=Gtk.Orientation.VERTICAL, spacing=8)
        host_entry = Adw.EntryRow(title=_("Hostname / IP Address"))
        host_entry.set_input_purpose(Gtk.InputPurpose.URL)
        port_entry = Adw.EntryRow(title=_("Port"))
        port_entry.set_input_purpose(Gtk.InputPurpose.DIGITS)
        port_entry.set_text("5000")
        fields_list = Gtk.ListBox(selection_mode=Gtk.SelectionMode.NONE)
        fields_list.add_css_class("boxed-list")
        fields_list.append(host_entry)
        fields_list.append(port_entry)
        content.append(fields_list)

        dialog.set_extra_child(content)
        dialog.add_response("cancel", _("Cancel"))
        dialog.add_response("connect", _("Connect"))
        dialog.set_response_appearance("connect", Adw.ResponseAppearance.SUGGESTED)

        def _on_response(d, response):
            if response != "connect":
                return
            host = host_entry.get_text().strip()
            port_str = port_entry.get_text().strip() or "5000"
            if not host:
                return
            try:
                port = int(port_str)
                if not 1 <= port <= 65535:
                    raise ValueError
            except ValueError:
                return
            from meshy.tcp import TcpDevice, TcpTransport

            transport = TcpTransport()
            device = TcpDevice(host, port)
            self._window.connect_to_device(device, transport=transport)

        dialog.connect("response", _on_response)
        dialog.present(self._window)

    def _start_bt_polling(self):
        """Poll for Bluetooth state changes every 2 seconds."""
        if self._bt_poll_id is not None:
            return

        def _check_bt():
            bt_enabled = self._window.ble_transport.is_bluetooth_enabled()
            if bt_enabled != self._bt_was_enabled:
                self.refresh_paired_devices()
            return True

        self._bt_poll_id = GLib.timeout_add(2000, _check_bt)

    def stop_polling(self):
        """Stop polling for Bluetooth state."""
        if self._bt_poll_id is not None:
            GLib.source_remove(self._bt_poll_id)
            self._bt_poll_id = None

    def _on_pair_new(self, *args):
        """Scan for new MeshCore devices and pair."""
        # Check if BLE transport is available
        if not self._window.ble_transport:
            return

        dialog = Adw.Dialog()
        dialog.set_title(_("Pair New Companion"))
        dialog.set_content_width(420)
        dialog.set_content_height(480)

        toolbar_view = Adw.ToolbarView()
        header = Adw.HeaderBar()
        toolbar_view.add_top_bar(header)

        content = Gtk.Box(
            orientation=Gtk.Orientation.VERTICAL,
            spacing=8,
            margin_start=8,
            margin_end=8,
            margin_bottom=8,
        )

        # Status with spinner
        status_box = Gtk.Box(spacing=8, halign=Gtk.Align.CENTER)
        spinner = Gtk.Spinner(spinning=True)
        status_box.append(spinner)
        status_label = Gtk.Label(label=_("Scanning for MeshCore devices..."))
        status_label.add_css_class("dim-label")
        status_box.append(status_label)
        content.append(status_box)

        # Device list
        scroll = Gtk.ScrolledWindow(vexpand=True)
        list_wrapper = Gtk.Box(
            orientation=Gtk.Orientation.VERTICAL,
            margin_start=8,
            margin_end=8,
            margin_top=8,
            margin_bottom=8,
        )
        list_box = Gtk.ListBox()
        list_box.set_selection_mode(Gtk.SelectionMode.NONE)
        list_box.add_css_class("boxed-list")

        placeholder = Adw.StatusPage(
            icon_name="bluetooth-active-symbolic",
            title=_("Scanning..."),
            description=_("Looking for nearby MeshCore devices"),
        )
        list_box.set_placeholder(placeholder)

        list_wrapper.append(list_box)
        scroll.set_child(list_wrapper)
        content.append(scroll)

        toolbar_view.set_content(content)
        dialog.set_child(toolbar_view)

        seen_addresses = set()
        paired_addresses = {
            d.address.upper() for d in self._window.ble_transport.get_paired_devices()
        }

        def on_device_found(device: BleDevice):
            if device.address in seen_addresses:
                return
            if device.address.upper() in paired_addresses:
                return
            seen_addresses.add(device.address)

            row = Adw.ActionRow(
                title=device.name,
                subtitle=device.address,
            )
            row.add_prefix(Gtk.Image.new_from_icon_name("bluetooth-active-symbolic"))
            pair_btn = Gtk.Button(label=_("Pair"), valign=Gtk.Align.CENTER)
            pair_btn.add_css_class("suggested-action")
            pair_btn.connect("clicked", lambda b, d=device: self._initiate_pairing(d, dialog))
            row.add_suffix(pair_btn)
            list_box.append(row)

        # Save original callback and set ours
        orig_callback = self._window.ble_transport.on_device_discovered
        self._window.ble_transport.on_device_discovered = on_device_found

        def on_dialog_closed(d):
            self._window.ble_transport.on_device_discovered = orig_callback
            if self._window.ble_transport.state.name == "SCANNING":
                self._window.ble_transport.stop_scan()
            # Refresh paired list in case a new device was just paired
            self.refresh_paired_devices()

        dialog.connect("closed", on_dialog_closed)

        # Start scanning
        self._window.ble_transport.start_scan()
        dialog.present(self._window)

    def _initiate_pairing(self, device: BleDevice, scan_dialog: Adw.Dialog):
        """Start pairing with a device, showing passkey dialog if needed."""
        # Stop scanning and close scan dialog
        if self._window.ble_transport.state.name == "SCANNING":
            self._window.ble_transport.stop_scan()
        scan_dialog.close()

        # Set up passkey handler
        def on_passkey_requested(device_path, reply_callback):
            GLib.idle_add(lambda: self._show_passkey_dialog(device_path, reply_callback) or False)

        self._window.ble_transport.on_passkey_requested = on_passkey_requested
        self._window.ble_transport.pair_device(device.path)

    def _show_passkey_dialog(self, device_path, reply_callback):
        """Show dialog asking user to enter the passkey displayed on the device."""
        dialog = Adw.AlertDialog(
            heading=_("Enter Pairing PIN"),
            body=_("Enter the PIN shown on your MeshCore device."),
        )

        content = Gtk.Box(orientation=Gtk.Orientation.VERTICAL, spacing=8)
        pin_entry = Adw.EntryRow(title=_("PIN"))
        pin_entry.set_input_purpose(Gtk.InputPurpose.DIGITS)
        fields_list = Gtk.ListBox(selection_mode=Gtk.SelectionMode.NONE)
        fields_list.add_css_class("boxed-list")
        fields_list.append(pin_entry)
        content.append(fields_list)
        dialog.set_extra_child(content)

        dialog.add_response("cancel", _("Cancel"))
        dialog.add_response("pair", _("Pair"))
        dialog.set_response_appearance("pair", Adw.ResponseAppearance.SUGGESTED)

        def _on_response(d, response):
            if response == "pair":
                try:
                    passkey = int(pin_entry.get_text().strip())
                    reply_callback(passkey)
                except ValueError:
                    reply_callback(0)
            else:
                reply_callback(0)

        dialog.connect("response", _on_response)
        dialog.present(self._window)



# Copyright 2026, Jiri Eischmann <jiri@eischmann.cz> and the meshy contributors
# SPDX-License-Identifier: GPL-3.0-or-later

"""XDG Background Portal integration for background mode and autostart."""

import logging
import os

from gi.repository import Gio, GLib

log = logging.getLogger(__name__)

PORTAL_BUS_NAME = "org.freedesktop.portal.Desktop"
PORTAL_OBJECT_PATH = "/org/freedesktop/portal/desktop"
PORTAL_BACKGROUND_IFACE = "org.freedesktop.portal.Background"
PORTAL_REQUEST_IFACE = "org.freedesktop.portal.Request"


class BackgroundPortal:
    def __init__(self):
        self._bus = None
        self._signal_id = None

    def _ensure_bus(self):
        if self._bus is None:
            try:
                self._bus = Gio.bus_get_sync(Gio.BusType.SESSION, None)
            except GLib.Error as e:
                log.warning("Cannot connect to session bus: %s", e.message)
                return False
        return True

    def request_background(self, reason, autostart, commandline, callback):
        """Request permission to run in background and optionally autostart.

        callback(bg_granted, autostart_granted) is called on the main thread.
        """
        if not self._ensure_bus():
            GLib.idle_add(lambda: callback(False, False) or False)
            return

        token = f"meshy_bg_{os.getpid()}"
        sender = self._bus.get_unique_name().replace(".", "_").lstrip(":")
        request_path = f"/org/freedesktop/portal/desktop/request/{sender}/{token}"

        self._cleanup_signal()
        self._signal_id = self._bus.signal_subscribe(
            PORTAL_BUS_NAME,
            PORTAL_REQUEST_IFACE,
            "Response",
            request_path,
            None,
            Gio.DBusSignalFlags.NO_MATCH_RULE,
            lambda *args: self._on_response(callback, *args),
        )

        options = {
            "handle_token": GLib.Variant("s", token),
            "reason": GLib.Variant("s", reason),
            "autostart": GLib.Variant("b", autostart),
        }
        if commandline:
            options["commandline"] = GLib.Variant("as", commandline)

        params = GLib.Variant("(sa{sv})", ("", options))

        self._bus.call(
            PORTAL_BUS_NAME,
            PORTAL_OBJECT_PATH,
            PORTAL_BACKGROUND_IFACE,
            "RequestBackground",
            params,
            None,
            Gio.DBusCallFlags.NONE,
            30000,
            None,
            self._on_call_done,
        )

    def _on_call_done(self, bus, result):
        try:
            bus.call_finish(result)
        except GLib.Error as e:
            log.warning(
                "Could not request background permission via portal: %s",
                e.message,
            )
            self._cleanup_signal()

    def _on_response(self, callback, bus, sender, path, iface, signal, params):
        self._cleanup_signal()
        response = params.get_child_value(0).get_uint32()
        if response != 0:
            log.info("Background permission denied (response=%d)", response)
            GLib.idle_add(lambda: callback(False, False) or False)
            return

        results = params.get_child_value(1)
        bg_val = results.lookup_value("background", GLib.VariantType.new("b"))
        autostart_val = results.lookup_value("autostart", GLib.VariantType.new("b"))
        bg_granted = bg_val.get_boolean() if bg_val else False
        autostart_granted = autostart_val.get_boolean() if autostart_val else False

        log.info(
            "Background portal: bg=%s autostart=%s",
            bg_granted,
            autostart_granted,
        )
        GLib.idle_add(lambda: callback(bg_granted, autostart_granted) or False)

    def set_status(self, message):
        if not self._ensure_bus():
            return
        try:
            options = {}
            if message:
                options["message"] = GLib.Variant("s", message)
            self._bus.call_sync(
                PORTAL_BUS_NAME,
                PORTAL_OBJECT_PATH,
                PORTAL_BACKGROUND_IFACE,
                "SetStatus",
                GLib.Variant("(a{sv})", (options,)),
                None,
                Gio.DBusCallFlags.NONE,
                5000,
                None,
            )
        except GLib.Error as e:
            log.warning("Could not set background status via portal: %s", e.message)

    def clear_status(self):
        self.set_status("")

    def _cleanup_signal(self):
        if self._signal_id and self._bus:
            self._bus.signal_unsubscribe(self._signal_id)
            self._signal_id = None



# Copyright 2026, Jiri Eischmann <jiri@eischmann.cz> and the meshy contributors
# SPDX-License-Identifier: GPL-3.0-or-later

"""CoreLocation-based location service for macOS."""

import logging

log = logging.getLogger(__name__)

try:
    from Foundation import NSObject
    from CoreLocation import (
        CLLocationManager,
        kCLAuthorizationStatusAuthorizedAlways,
        kCLAuthorizationStatusAuthorizedWhenInUse,
        kCLAuthorizationStatusDenied,
        kCLAuthorizationStatusNotDetermined,
        kCLAuthorizationStatusRestricted,
    )
    CORELOCATION_AVAILABLE = True
except ImportError:
    log.warning("CoreLocation not available")
    CORELOCATION_AVAILABLE = False
    NSObject = object


class LocationDelegate(NSObject):
    """Delegate for CoreLocation manager."""

    def init(self):
        """Initialize the delegate."""
        from objc import super as objc_super
        self = objc_super(LocationDelegate, self).init()
        if self is None:
            return None

        self._callback = None
        self._manager = None
        return self

    def setCallback_(self, callback):
        """Set the callback to call with location results."""
        self._callback = callback

    def setManager_(self, manager):
        """Set the location manager."""
        self._manager = manager

    def locationManager_didUpdateLocations_(self, manager, locations):
        """Called when location is updated."""
        if not locations or not self._callback:
            return

        location = locations[-1]  # Most recent location
        coordinate = location.coordinate()

        log.info(f"Location updated: {coordinate.latitude}, {coordinate.longitude}")

        # Call callback on main thread
        from gi.repository import GLib
        GLib.idle_add(
            lambda: self._callback(coordinate.latitude, coordinate.longitude, None)
            if self._callback else None
        )

        # Stop updating after getting one location
        if self._manager:
            self._manager.stopUpdatingLocation()

    def locationManager_didFailWithError_(self, manager, error):
        """Called when location update fails."""
        log.error(f"Location error: {error}")

        # Call callback with error on main thread
        from gi.repository import GLib
        GLib.idle_add(
            lambda: self._callback(None, None, str(error))
            if self._callback else None
        )

    def locationManager_didChangeAuthorizationStatus_(self, manager, status):
        """Called when authorization status changes."""
        if status == kCLAuthorizationStatusAuthorizedWhenInUse:
            log.info("Location permission granted (when in use)")
            manager.requestLocation()
        elif status == kCLAuthorizationStatusAuthorizedAlways:
            log.info("Location permission granted (always)")
            manager.requestLocation()
        elif status == kCLAuthorizationStatusDenied:
            log.warning("Location permission denied")
            if self._callback:
                from gi.repository import GLib
                GLib.idle_add(
                    lambda: self._callback(None, None, "Location permission denied")
                )
        elif status == kCLAuthorizationStatusRestricted:
            log.warning("Location permission restricted")
            if self._callback:
                from gi.repository import GLib
                GLib.idle_add(
                    lambda: self._callback(None, None, "Location permission restricted")
                )


def get_location_macos(callback):
    """Get current location using CoreLocation.

    Args:
        callback: Function to call with (latitude, longitude, error).
                  If successful, error is None. If failed, lat/lon are None.
    """
    if not CORELOCATION_AVAILABLE:
        from gi.repository import GLib
        GLib.idle_add(lambda: callback(None, None, "CoreLocation not available"))
        return

    try:
        # Create location manager and delegate on main thread
        from gi.repository import GLib

        def _init_location():
            manager = CLLocationManager.alloc().init()
            delegate = LocationDelegate.alloc().init()

            delegate.setCallback_(callback)
            delegate.setManager_(manager)
            manager.setDelegate_(delegate)

            # Request permission
            auth_status = CLLocationManager.authorizationStatus()

            if auth_status == kCLAuthorizationStatusNotDetermined:
                log.info("Requesting location permission")
                manager.requestWhenInUseAuthorization()
            elif auth_status in (kCLAuthorizationStatusAuthorizedWhenInUse,
                                 kCLAuthorizationStatusAuthorizedAlways):
                log.info("Location already authorized, requesting location")
                manager.requestLocation()
            elif auth_status == kCLAuthorizationStatusDenied:
                callback(None, None, "Location permission denied")
            elif auth_status == kCLAuthorizationStatusRestricted:
                callback(None, None, "Location permission restricted")

            return False  # Don't repeat

        GLib.idle_add(_init_location)

    except Exception as e:
        log.error(f"Failed to get location: {e}")
        from gi.repository import GLib
        GLib.idle_add(lambda: callback(None, None, str(e)))



# Copyright 2026, Jiri Eischmann <jiri@eischmann.cz> and the meshy contributors
# SPDX-License-Identifier: GPL-3.0-or-later

"""SMAZ decompression for interoperability with meshcore-open clients.

Meshcore-open can compress short text messages using the SMAZ algorithm
(dictionary-based compression for short ASCII strings). Compressed messages
are sent with an 's:' prefix followed by base64-encoded SMAZ data.
"""

import base64

_VERBATIM_SINGLE = 254
_VERBATIM_RUN = 255

_CODEBOOK = [
    " ",
    "the",
    "e",
    "t",
    "a",
    "of",
    "o",
    "and",
    "i",
    "n",
    "s",
    "e ",
    "r",
    " th",
    " t",
    "in",
    "he",
    "th",
    "h",
    "he ",
    "to",
    "\r\n",
    "l",
    "s ",
    "d",
    " a",
    "an",
    "er",
    "c",
    " o",
    "d ",
    "on",
    " of",
    "re",
    "of ",
    "t ",
    ", ",
    "is",
    "u",
    "at",
    "   ",
    "n ",
    "or",
    "which",
    "f",
    "m",
    "as",
    "it",
    "that",
    "\n",
    "was",
    "en",
    "  ",
    " w",
    "es",
    " an",
    " i",
    "\r",
    "f ",
    "g",
    "p",
    "nd",
    " s",
    "nd ",
    "ed ",
    "w",
    "ed",
    "http://",
    "for",
    "te",
    "ing",
    "y ",
    "The",
    " c",
    "ti",
    "r ",
    "his",
    "st",
    " in",
    "ar",
    "nt",
    ",",
    " to",
    "y",
    "ng",
    " h",
    "with",
    "le",
    "al",
    "to ",
    "b",
    "ou",
    "be",
    "were",
    " b",
    "se",
    "o ",
    "ent",
    "ha",
    "ng ",
    "their",
    '"',
    "hi",
    "from",
    " f",
    "in ",
    "de",
    "ion",
    "me",
    "v",
    ".",
    "ve",
    "all",
    "re ",
    "ri",
    "ro",
    "is ",
    "co",
    "f t",
    "are",
    "ea",
    ". ",
    "her",
    " m",
    "er ",
    " p",
    "es ",
    "by",
    "they",
    "di",
    "ra",
    "ic",
    "not",
    "s, ",
    "d t",
    "at ",
    "ce",
    "la",
    "h ",
    "ne",
    "as ",
    "tio",
    "on ",
    "n t",
    "io",
    "we",
    " a ",
    "om",
    ", a",
    "s o",
    "ur",
    "li",
    "ll",
    "ch",
    "had",
    "this",
    "e t",
    "g ",
    "e\r\n",
    " wh",
    "ere",
    " co",
    "e o",
    "a ",
    "us",
    " d",
    "ss",
    "\n\r\n",
    "\r\n\r",
    '="',
    " be",
    " e",
    "s a",
    "ma",
    "one",
    "t t",
    "or ",
    "but",
    "el",
    "so",
    "l ",
    "e s",
    "s,",
    "no",
    "ter",
    " wa",
    "iv",
    "ho",
    "e a",
    " r",
    "hat",
    "s t",
    "ns",
    "ch ",
    "wh",
    "tr",
    "ut",
    "/",
    "have",
    "ly ",
    "ta",
    " ha",
    " on",
    "tha",
    "-",
    " l",
    "ati",
    "en ",
    "pe",
    " re",
    "there",
    "ass",
    "si",
    " fo",
    "wa",
    "ec",
    "our",
    "who",
    "its",
    "z",
    "fo",
    "rs",
    ">",
    "ot",
    "un",
    "<",
    "im",
    "th ",
    "nc",
    "ate",
    "><",
    "ver",
    "ad",
    " we",
    "ly",
    "ee",
    " n",
    "id",
    " cl",
    "ac",
    "il",
    "</",
    "rt",
    " wi",
    "div",
    "e, ",
    " it",
    "whi",
    " ma",
    "ge",
    "x",
    "e c",
    "men",
    ".com",
]

_CODEBOOK_BYTES = [entry.encode("ascii") for entry in _CODEBOOK]


def _decompress(data: bytes) -> bytes:
    out = bytearray()
    i = 0
    while i < len(data):
        code = data[i]
        if code == _VERBATIM_SINGLE:
            if i + 1 >= len(data):
                break
            out.append(data[i + 1])
            i += 2
        elif code == _VERBATIM_RUN:
            if i + 1 >= len(data):
                break
            length = data[i + 1] + 1
            end = i + 2 + length
            if end > len(data):
                break
            out.extend(data[i + 2 : end])
            i = end
        elif code < len(_CODEBOOK_BYTES):
            out.extend(_CODEBOOK_BYTES[code])
            i += 1
        else:
            break
    return bytes(out)


def try_decode_prefixed(text: str) -> str:
    """If text starts with 's:', decode base64 + SMAZ. Otherwise return as-is."""
    stripped = text.lstrip()
    if not stripped.startswith("s:") or len(stripped) <= 2:
        return text
    encoded = stripped[2:]
    try:
        compressed = base64.b64decode(encoded, validate=True)
    except Exception:
        try:
            normalized = encoded.replace("-", "+").replace("_", "/")
            pad = len(normalized) % 4
            if pad:
                normalized += "=" * (4 - pad)
            compressed = base64.b64decode(normalized)
        except Exception:
            return text
    try:
        return _decompress(compressed).decode("utf-8", errors="replace")
    except Exception:
        return text



import base64

from meshy.smaz import _decompress, try_decode_prefixed


class TestDecompress:
    def test_codebook_lookup(self):
        result = _decompress(bytes([0]))
        assert result == b" "

    def test_codebook_the(self):
        result = _decompress(bytes([1]))
        assert result == b"the"

    def test_verbatim_single(self):
        result = _decompress(bytes([254, ord("X")]))
        assert result == b"X"

    def test_verbatim_run(self):
        result = _decompress(bytes([255, 2, ord("A"), ord("B"), ord("C")]))
        assert result == b"ABC"

    def test_mixed(self):
        result = _decompress(bytes([1, 0, 2]))
        assert result == b"the e"

    def test_empty(self):
        assert _decompress(b"") == b""


class TestTryDecodePrefixed:
    def test_no_prefix_returns_original(self):
        assert try_decode_prefixed("hello world") == "hello world"

    def test_plain_s_prefix_with_valid_smaz(self):
        compressed = bytes([1])  # "the"
        encoded = base64.b64encode(compressed).decode()
        result = try_decode_prefixed(f"s:{encoded}")
        assert result == "the"

    def test_non_base64_chars_stripped_by_lenient_decode(self):
        result = try_decode_prefixed("s:@@@")
        assert result == ""

    def test_url_safe_base64(self):
        compressed = bytes([254, ord("Z")])
        encoded = base64.urlsafe_b64encode(compressed).decode().rstrip("=")
        result = try_decode_prefixed(f"s:{encoded}")
        assert result == "Z"

    def test_empty_after_prefix_returns_original(self):
        assert try_decode_prefixed("s:") == "s:"



# Copyright 2026, Jiri Eischmann <jiri@eischmann.cz> and the meshy contributors
# SPDX-License-Identifier: GPL-3.0-or-later

"""Shared view utilities."""

import json
import logging
import threading
import weakref
from datetime import datetime, timedelta
from urllib.request import Request, urlopen

import gi

gi.require_version("Adw", "1")
gi.require_version("Gtk", "4.0")
gi.require_version("Shumate", "1.0")
from gi.repository import Adw, GLib, GObject, Gtk, Shumate

_OFM_BASE = "https://tiles.openfreemap.org"
_OFM_STYLE_URLS = {
    "bright": f"{_OFM_BASE}/styles/bright",
    "dark": f"{_OFM_BASE}/styles/dark",
}

log = logging.getLogger(__name__)

_source_cache = {}
_source_failed = set()
_map_widgets = []
_theme_listener_connected = False
_custom_tile_url = None
_custom_tile_type = None

_UA = "Meshy/1.0 (https://codeberg.org/sesivany/meshy)"
_MAX_ZOOM = 14
_RASTER_MAX_ZOOM = 19


def _format_date_text(dt):
    """Format a datetime as a human-friendly date string for dividers."""
    today = datetime.now().date()
    d = dt.date() if isinstance(dt, datetime) else dt
    if d == today:
        return "Today"
    if d == today - timedelta(days=1):
        return "Yesterday"
    if d.year == today.year:
        return d.strftime("%A, %B %-d")
    return d.strftime("%A, %B %-d, %Y")


_osm_source = None


def _get_osm_source():
    global _osm_source
    if _osm_source is None:
        registry = Shumate.MapSourceRegistry.new_with_defaults()
        _osm_source = registry.get_by_id(Shumate.MAP_SOURCE_OSM_MAPNIK)
    return _osm_source


def _make_raster_source(url_template):
    return Shumate.RasterRenderer.new_full_from_url(
        "custom-tiles",
        "Custom Tiles",
        "",
        "",
        0,
        19,
        256,
        Shumate.MapProjection.MERCATOR,
        url_template,
    )


def _fix_interpolations(obj):
    """Fix ["linear", N] -> ["linear"] for Shumate compatibility."""
    if isinstance(obj, list):
        if (
            len(obj) >= 2
            and obj[0] == "interpolate"
            and isinstance(obj[1], list)
            and obj[1][0] == "linear"
            and len(obj[1]) > 1
        ):
            obj[1] = ["linear"]
        for item in obj:
            _fix_interpolations(item)
    elif isinstance(obj, dict):
        for v in obj.values():
            _fix_interpolations(v)


def _patch_style_json(style_text):
    """Patch a vector style JSON for Shumate compatibility.

    Shumate supports only one data source. When the style has multiple vector
    sources, we keep the one referenced by the most layers (the "main" source)
    and drop layers that reference other sources.
    """
    from collections import Counter

    style = json.loads(style_text)
    sources = style.get("sources", {})
    vector_sources = {k: s for k, s in sources.items() if s.get("type") == "vector"}
    if not vector_sources:
        return None
    layer_counts = Counter(
        ly.get("source") for ly in style.get("layers", []) if ly.get("source") in vector_sources
    )
    vector_key = layer_counts.most_common(1)[0][0] if layer_counts else next(iter(vector_sources))
    vector_src = vector_sources[vector_key]
    if "url" in vector_src and "tiles" not in vector_src:
        tilejson_text = (
            urlopen(Request(vector_src["url"], headers={"User-Agent": _UA}), timeout=10)
            .read()
            .decode("utf-8")
        )
        vector_src["tiles"] = json.loads(tilejson_text).get("tiles", [])
    vector_src.pop("url", None)
    style["sources"] = {vector_key: vector_src}
    _UNSUPPORTED_LAYER_TYPES = {"fill-extrusion", "heatmap", "hillshade", "raster"}
    style["layers"] = [
        ly
        for ly in style.get("layers", [])
        if (ly.get("source") in (vector_key, None) or "source" not in ly)
        and ly.get("type") not in _UNSUPPORTED_LAYER_TYPES
    ]
    style.pop("sprite", None)
    style.pop("glyphs", None)
    for layer in style["layers"]:
        _fix_interpolations(layer)
        _fix_offset_expressions(layer)
    return json.dumps(style)


def _is_simple_offset(val):
    return isinstance(val, list) and len(val) == 2 and all(isinstance(v, int | float) for v in val)


def _fix_offset_expressions(layer):
    """Replace expression-based offsets with [0, 0] — Shumate expects plain arrays."""
    for section_key in ("layout", "paint"):
        section = layer.get(section_key)
        if not section:
            continue
        for prop in ("text-offset", "icon-offset"):
            if prop in section and not _is_simple_offset(section[prop]):
                section[prop] = [0, 0]


def _fetch_vector_source(
    style_url, renderer_id, callback, cache_key=None, license_text=None, license_uri=None
):
    """Fetch a vector style JSON and create a VectorRenderer asynchronously."""
    log.warning(
        "fetch_vector_source: url=%s renderer_id=%s cache_key=%s", style_url, renderer_id, cache_key
    )
    if cache_key and cache_key in _source_cache:
        log.warning("fetch_vector_source: cache hit for %s", cache_key)
        callback(_source_cache[cache_key])
        return
    if cache_key and cache_key in _source_failed:
        log.warning("fetch_vector_source: previously failed for %s", cache_key)
        callback(None)
        return

    def _worker():
        try:
            log.warning("fetch_vector_source: fetching style from %s", style_url)
            style_text = (
                urlopen(Request(style_url, headers={"User-Agent": _UA}), timeout=15)
                .read()
                .decode("utf-8")
            )
            log.warning("fetch_vector_source: fetched %d bytes, patching...", len(style_text))
            patched = _patch_style_json(style_text)
            if not patched:
                log.warning("No vector source in style from %s", style_url)
                if cache_key:
                    _source_failed.add(cache_key)
                GLib.idle_add(lambda: callback(None) or False)
                return

            log.warning(
                "fetch_vector_source: patched to %d bytes, creating renderer...", len(patched)
            )

            def _finish():
                try:
                    renderer = Shumate.VectorRenderer.new(renderer_id, patched)
                    log.warning("fetch_vector_source: VectorRenderer created: %s", renderer)
                    if license_text:
                        renderer.set_license(license_text)
                    if license_uri:
                        renderer.set_license_uri(license_uri)
                    if cache_key:
                        _source_cache[cache_key] = renderer
                    callback(renderer)
                except Exception as e:
                    log.warning("Failed to create VectorRenderer %s: %s", renderer_id, e)
                    if cache_key:
                        _source_failed.add(cache_key)
                    callback(None)
                return False

            GLib.idle_add(_finish)
        except Exception as e:
            log.warning("Failed to fetch vector style from %s: %s", style_url, e)
            if cache_key:
                _source_failed.add(cache_key)
            GLib.idle_add(lambda: callback(None) or False)

    threading.Thread(target=_worker, daemon=True).start()


def _fetch_ofm_source(style, callback):
    """Fetch an OFM vector source (bright/dark) with caching."""
    _fetch_vector_source(
        _OFM_STYLE_URLS[style],
        f"ofm-{style}",
        callback,
        cache_key=style,
        license_text="© OpenFreeMap © OpenStreetMap (ODbL)",
        license_uri="https://www.openstreetmap.org/copyright",
    )


def _collect_live_widgets():
    """Prune dead weak refs and return live map widgets."""
    live = [(ref, ref()) for ref in _map_widgets]
    _map_widgets[:] = [ref for ref, w in live if w is not None]
    return [w for _, w in live if w is not None]


def _apply_map_source(map_widget, source):
    if source:
        map_widget.set_map_source(source)
        max_zoom = _RASTER_MAX_ZOOM if _custom_tile_type == "raster" else _MAX_ZOOM
        map_widget.get_viewport().set_max_zoom_level(max_zoom)


def _is_dark():
    return Adw.StyleManager.get_default().get_dark()


def _apply_default_source(widgets):
    """Apply the theme-appropriate OFM source (or OSM fallback) to widgets."""
    style = "dark" if _is_dark() else "bright"

    def _apply(source):
        s = source if source else _get_osm_source()
        for w in widgets:
            _apply_map_source(w, s)

    _fetch_ofm_source(style, _apply)


def _on_theme_changed(*_args):
    if _custom_tile_url:
        return
    widgets = _collect_live_widgets()
    if widgets:
        _apply_default_source(widgets)


def create_shumate_map(zoom_level=14, lat=0.0, lon=0.0):
    """Create a Shumate SimpleMap with theme-aware tiles.

    Raises:
        Exception: If map creation fails (e.g., on macOS with GdkMacos backend issues)
    """
    global _theme_listener_connected, _custom_tile_url, _custom_tile_type

    import logging

    log = logging.getLogger(__name__)

    try:
        log.debug(f"Creating Shumate map: zoom={zoom_level}, lat={lat}, lon={lon}")

        from gi.repository import Gio

        settings = Gio.Settings.new("page.codeberg.sesivany.Meshy")
        _custom_tile_url = settings.get_string("custom-tile-url") or None
        _custom_tile_type = settings.get_string("custom-tile-type") or None

        map_widget = Shumate.SimpleMap()
        map_widget.set_vexpand(True)

        viewport = map_widget.get_viewport()
        viewport.set_zoom_level(zoom_level)
        viewport.set_location(lat, lon)

        _collect_live_widgets()
        _map_widgets.append(weakref.ref(map_widget))

        # Apply tile source BEFORE adding marker layer so markers render on top
        if _custom_tile_url and _custom_tile_type == "raster":
            _apply_map_source(map_widget, _make_raster_source(_custom_tile_url))
        else:
            osm_source = _get_osm_source()
            if osm_source:
                _apply_map_source(map_widget, osm_source)

        marker_layer = Shumate.MarkerLayer.new(viewport)
        map_widget.get_map().add_layer(marker_layer)

        def _apply_async(source):
            if source:
                _apply_map_source(map_widget, source)

        if _custom_tile_url and _custom_tile_type == "vector":
            _fetch_vector_source(_custom_tile_url, "custom-vector", _apply_async)
        elif not _custom_tile_url:
            _apply_default_source([map_widget])

        if not _theme_listener_connected:
            Adw.StyleManager.get_default().connect("notify::dark", _on_theme_changed)
            _theme_listener_connected = True

        log.debug("Shumate map created successfully")
        return map_widget, marker_layer, viewport

    except Exception as e:
        log.error(f"Failed to create Shumate map widget: {e}", exc_info=True)
        # Re-raise so callers can handle gracefully
        raise


def apply_custom_source(url, tile_type, on_result=None):
    """Switch all tracked map widgets to a custom source, or back to default.

    on_result(ok: bool) is called on the main thread when done (vector only).
    """
    global _custom_tile_url, _custom_tile_type
    _custom_tile_url = url or None
    _custom_tile_type = tile_type or None

    widgets = _collect_live_widgets()
    log.warning("apply_custom_source: url=%s type=%s widgets=%d", url, tile_type, len(widgets))
    if not widgets:
        log.warning("apply_custom_source: no live widgets — source saved for next map open")
        if on_result:
            on_result(True)
        return

    if _custom_tile_url and _custom_tile_type == "raster":
        for w in widgets:
            _apply_map_source(w, _make_raster_source(_custom_tile_url))
        if on_result:
            on_result(True)
    elif _custom_tile_url and _custom_tile_type == "vector":

        def _apply(source):
            log.warning("apply_custom_source callback: source=%s widgets=%d", source, len(widgets))
            if source:
                for w in widgets:
                    _apply_map_source(w, source)
            if on_result:
                on_result(source is not None)

        _fetch_vector_source(_custom_tile_url, "custom-vector", _apply)
    else:
        _apply_default_source(widgets)
        if on_result:
            on_result(True)


def deregister_map_widget(map_widget):
    """Remove a map widget from the theme-tracking list."""
    _map_widgets[:] = [ref for ref in _map_widgets if ref() is not None and ref() is not map_widget]


def open_map_picker(window, title, button_label, init_lat, init_lon, on_pick):
    """Open a map dialog with crosshair, call on_pick(lat, lon) on confirm."""
    dialog = Adw.Dialog()
    dialog.set_title(title)
    win_w = window.get_width() or 800
    win_h = window.get_height() or 600
    dialog.set_content_width(max(500, win_w - 60))
    dialog.set_content_height(max(400, win_h - 60))

    toolbar_view = Adw.ToolbarView()
    header = Adw.HeaderBar()
    confirm_btn = Gtk.Button(label=button_label)
    confirm_btn.add_css_class("suggested-action")
    header.pack_end(confirm_btn)
    toolbar_view.add_top_bar(header)

    init_zoom = 14 if (abs(init_lat) > 1e-6 or abs(init_lon) > 1e-6) else 2
    map_widget, _ml, viewport = create_shumate_map(zoom_level=init_zoom, lat=init_lat, lon=init_lon)
    map_widget.set_hexpand(True)

    crosshair = Gtk.Label()
    crosshair.set_markup('<span size="24000" color="#e05050">✚</span>')
    crosshair.set_halign(Gtk.Align.CENTER)
    crosshair.set_valign(Gtk.Align.CENTER)

    overlay = Gtk.Overlay()
    overlay.set_child(map_widget)
    overlay.add_overlay(crosshair)

    toolbar_view.set_content(overlay)
    dialog.set_child(toolbar_view)

    def _on_confirm(*_args):
        on_pick(viewport.get_latitude(), viewport.get_longitude())
        dialog.close()

    confirm_btn.connect("clicked", _on_confirm)
    dialog.connect("closed", lambda _d: deregister_map_widget(map_widget))
    dialog.present(window)


def _add_hover_swap(row, button, badge, get_count, on_long_press):
    """Add hover show/hide behavior: show button and hide badge on hover,
    reverse on leave. Also add long-press handler for touchscreens."""
    motion = Gtk.EventControllerMotion()
    motion.connect("enter", lambda *_: (button.set_visible(True), badge.set_visible(False)))
    motion.connect(
        "leave", lambda *_: (button.set_visible(False), badge.set_visible(get_count() > 0))
    )
    row.add_controller(motion)
    long_press = Gtk.GestureLongPress()
    long_press.connect("pressed", lambda g, x, y: on_long_press())
    row.add_controller(long_press)


_CONTACT_TYPE_ICONS = {
    1: "avatar-default-symbolic",
    2: "network-server-symbolic",
    3: "user-home-symbolic",
    4: "weather-clear-symbolic",
}


class _ContactPickItem(GObject.Object):
    def __init__(self, contact):
        super().__init__()
        self.contact = contact


def _contact_pick_factory_setup(factory, list_item, button_label):
    row = Adw.ActionRow()
    icon = Gtk.Image()
    row.add_prefix(icon)
    btn = Gtk.Button(label=button_label, valign=Gtk.Align.CENTER)
    btn.add_css_class("suggested-action")
    btn.set_visible(False)
    row.add_suffix(btn)
    row._icon = icon
    row._action_btn = btn
    row._action_handler_id = None

    hover = Gtk.EventControllerMotion()
    hover.connect("enter", lambda *_a: btn.set_visible(True))
    hover.connect("leave", lambda *_a: btn.set_visible(False))
    row.add_controller(hover)

    list_item.set_child(row)


def _contact_pick_factory_bind(factory, list_item, on_click, subtitle_func=None):
    row = list_item.get_child()
    contact = list_item.get_item().contact

    row.set_title(GLib.markup_escape_text(contact.name) if contact.name else _("Unknown"))
    subtitle = subtitle_func(contact) if subtitle_func else contact.type_label
    row.set_subtitle(subtitle)
    row._icon.set_from_icon_name(_CONTACT_TYPE_ICONS.get(contact.type, "avatar-default-symbolic"))

    row._action_handler_id = row._action_btn.connect("clicked", lambda _b: on_click(contact))


def _contact_pick_factory_unbind(factory, list_item):
    row = list_item.get_child()
    if row._action_handler_id is not None:
        row._action_btn.disconnect(row._action_handler_id)
        row._action_handler_id = None



# Copyright 2026, Jiri Eischmann <jiri@eischmann.cz> and the meshy contributors
# SPDX-License-Identifier: GPL-3.0-or-later

"""Lightweight MeshCore packet decoding and channel crypto.

Replaces meshcoredecoder for the subset of features Meshy needs:
- Channel hash calculation (SHA-256)
- GroupText payload decryption (HMAC-SHA256 + AES-128-ECB)
- Minimal packet header/path parsing for LOG_DATA frames
"""

import hashlib
import hmac
import struct

try:
    from Crypto.Cipher import AES
except ImportError:
    AES = None

# Payload type constants (from header bits 2-5)
PAYLOAD_ADVERT = 0x04
PAYLOAD_GROUP_TEXT = 0x05


def channel_hash(psk_hex: str) -> str:
    """Calculate the 1-byte channel hash from a PSK hex string.

    This is SHA-256(psk_bytes)[0] as a 2-char lowercase hex string.
    """
    h = hashlib.sha256(bytes.fromhex(psk_hex)).digest()
    return f"{h[0]:02x}"


def decrypt_group_text(ciphertext: bytes, psk: bytes) -> dict | None:
    """Decrypt a GroupText payload using the channel PSK.

    Args:
        ciphertext: The encrypted payload (after channel_hash byte).
                    First 2 bytes are HMAC-SHA256 MAC, rest is AES-128-ECB ciphertext.
        psk: The 16-byte channel pre-shared key.

    Returns:
        Dict with 'sender', 'message', 'timestamp', 'flags' or None on failure.
    """
    if len(ciphertext) < 3:
        return None

    mac = ciphertext[:2]
    encrypted = ciphertext[2:]

    # Verify HMAC-SHA256: key is PSK padded to 32 bytes with zeros
    key32 = psk + b"\x00" * (32 - len(psk)) if len(psk) < 32 else psk[:32]
    expected_mac = hmac.new(key32, encrypted, hashlib.sha256).digest()
    if mac[0] != expected_mac[0] or mac[1] != expected_mac[1]:
        return None

    # Decrypt with AES-128-ECB
    if AES is None:
        return None
    key16 = psk[:16] if len(psk) >= 16 else psk + b"\x00" * (16 - len(psk))
    cipher = AES.new(key16, AES.MODE_ECB)
    plaintext = cipher.decrypt(encrypted)

    # Parse plaintext: timestamp(4 LE) + flags(1) + "sender: message"
    if len(plaintext) < 6:
        return None

    timestamp = struct.unpack("<I", plaintext[:4])[0]
    flags = plaintext[4]
    text = plaintext[5:].decode("utf-8", "ignore").rstrip("\x00")

    # Split "sender: message" on first ": "
    if ": " in text:
        sender, message = text.split(": ", 1)
    else:
        sender = ""
        message = text

    return {
        "sender": sender,
        "message": message,
        "timestamp": timestamp,
        "flags": flags,
    }


def decode_log_packet(raw: bytes) -> dict | None:
    """Parse a raw mesh packet from LOG_DATA, extract GroupText if present.

    Parses the MeshCore packet header, skips path data, and extracts
    the payload. Only GroupText payloads are fully decoded.

    Args:
        raw: Raw packet bytes (after LOG_DATA framing).

    Returns:
        Dict with 'payload_type', 'path' (list of hex hop strings),
        'channel_hash', and 'decrypted' (if GroupText), or None.
    """
    if len(raw) < 2:
        return None

    header = raw[0]
    route_type = header & 0x03
    payload_type = (header >> 2) & 0x0F

    offset = 1

    # Parse transport codes for TC_FLOOD (0) and TC_DIRECT (3)
    transport_from = ""
    transport_to = ""
    if route_type == 0 or route_type == 3:
        if len(raw) < offset + 4:
            return None
        transport_from = raw[offset : offset + 2].hex()
        transport_to = raw[offset + 2 : offset + 4].hex()
        offset += 4

    # Parse path length byte
    if len(raw) < offset + 1:
        return None
    path_len_byte = raw[offset]
    offset += 1

    # Decode path: bits 7:6 = hash_size selector, bits 5:0 = hop count
    hash_size = (path_len_byte >> 6) + 1
    hop_count = path_len_byte & 63
    path_byte_length = hop_count * hash_size

    # Legacy fallback: reserved hash_size (4) or path extends past packet
    if hash_size == 4 or offset + path_byte_length > len(raw):
        path_byte_length = min(path_len_byte, len(raw) - offset, 64)
        hop_count = path_byte_length
        hash_size = 1

    # Extract path hops as hex strings
    path = []
    for i in range(hop_count):
        hop_start = offset + i * hash_size
        hop_bytes = raw[hop_start : hop_start + hash_size]
        path.append(hop_bytes.hex())
    offset += path_byte_length

    # Extract payload
    payload = raw[offset:]

    pkt_hash = hashlib.sha256(bytes([payload_type]) + payload).hexdigest()[:16]

    result = {
        "payload_type": payload_type,
        "route_type": route_type,
        "path": path,
        "path_hash_size": hash_size,
        "pkt_hash": pkt_hash,
        "transport_from": transport_from,
        "transport_to": transport_to,
    }

    if payload_type == PAYLOAD_ADVERT:
        advert = parse_advert_payload(payload)
        if advert:
            result["advert"] = advert

    # Decode GroupText payload: channel_hash(1) + mac(2) + ciphertext
    elif payload_type == PAYLOAD_GROUP_TEXT and len(payload) >= 3:
        result["channel_hash"] = f"{payload[0]:02x}"
        result["ciphertext"] = payload[1:]  # mac + encrypted data

    return result


def parse_advert_payload(payload: bytes) -> dict | None:
    """Parse an ADVERT payload from a LOG_DATA frame.

    Format: pubkey(32) + timestamp(4) + signature(64) + flags(1)
            + [lat(4) + lon(4) if flag 0x10] + [name(null-terminated) if flag 0x80]
    """
    if len(payload) < 101:  # 32 + 4 + 64 + 1
        return None
    pub_key = payload[:32].hex()
    timestamp = struct.unpack_from("<I", payload, 32)[0]
    # Skip 64-byte signature
    flags = payload[100]
    node_type = flags & 0x0F
    has_location = bool(flags & 0x10)
    has_name = bool(flags & 0x80)

    offset = 101
    lat, lon = None, None
    if has_location and len(payload) >= offset + 8:
        lat = struct.unpack_from("<i", payload, offset)[0] / 1e6
        lon = struct.unpack_from("<i", payload, offset + 4)[0] / 1e6
        if abs(lat) < 1e-6 and abs(lon) < 1e-6:
            lat, lon = None, None
        offset += 8

    name = ""
    if has_name and len(payload) > offset:
        end = payload.index(0, offset) if 0 in payload[offset:] else len(payload)
        name = payload[offset:end].decode("utf-8", "ignore")

    return {
        "public_key": pub_key,
        "type": node_type,
        "adv_name": name,
        "last_advert": timestamp,
        "adv_lat": lat,
        "adv_lon": lon,
    }



# Copyright 2026, Jiri Eischmann <jiri@eischmann.cz> and the meshy contributors
# SPDX-License-Identifier: GPL-3.0-or-later

"""Data models for MeshCore entities."""

import hashlib
from dataclasses import dataclass, field
from datetime import datetime
from enum import IntEnum, auto


def battery_percent_from_mv(mv: int) -> int:
    """Estimate LiPo battery percentage from voltage.

    Piecewise linear approximation with breakpoints at
    4.2V/100%, 4.0V/90%, 3.6V/10%, 3.5V/5%, 3.0V/0%.
    """
    _TABLE = [(4200, 100), (4000, 90), (3600, 10), (3500, 5), (3000, 0)]
    if mv >= _TABLE[0][0]:
        return 100
    if mv <= _TABLE[-1][0]:
        return 0
    for i in range(len(_TABLE) - 1):
        if mv >= _TABLE[i + 1][0]:
            v_hi, p_hi = _TABLE[i]
            v_lo, p_lo = _TABLE[i + 1]
            return round(p_lo + (p_hi - p_lo) * (mv - v_lo) / (v_hi - v_lo))
    return 0


def battery_icon_name(percent: int) -> str:
    if percent >= 90:
        return "battery-level-100-symbolic"
    if percent >= 80:
        return "battery-level-80-symbolic"
    if percent >= 70:
        return "battery-level-70-symbolic"
    if percent >= 60:
        return "battery-level-60-symbolic"
    if percent >= 40:
        return "battery-level-50-symbolic"
    if percent >= 30:
        return "battery-level-30-symbolic"
    if percent >= 20:
        return "battery-level-20-symbolic"
    return "battery-level-low-symbolic"


SNR_THRESHOLDS = {
    7: (4.0, -2.0, -4.0, -6.0),
    8: (4.0, -4.0, -6.0, -8.0),
    9: (4.0, -6.0, -8.0, -10.0),
    10: (4.0, -8.0, -10.0, -13.0),
    11: (4.0, -10.0, -12.5, -15.0),
    12: (4.0, -12.5, -15.0, -18.0),
}

SIGNAL_ICONS = (
    "strength-bars-1-symbolic",
    "strength-bars-2-symbolic",
    "strength-bars-3-symbolic",
    "strength-bars-4-symbolic",
    "strength-bars-5-symbolic",
)


SIGNAL_LABELS = (
    "Excellent",
    "Good",
    "Fair",
    "Poor",
    "Very poor",
)


def snr_to_icon(snr: float, sf: int = 12) -> str:
    thresholds = SNR_THRESHOLDS.get(sf, SNR_THRESHOLDS[12])
    for i, t in enumerate(thresholds):
        if snr >= t:
            return SIGNAL_ICONS[i]
    return SIGNAL_ICONS[4]


def snr_to_label(snr: float, sf: int = 12) -> str:
    thresholds = SNR_THRESHOLDS.get(sf, SNR_THRESHOLDS[12])
    for i, t in enumerate(thresholds):
        if snr >= t:
            return SIGNAL_LABELS[i]
    return SIGNAL_LABELS[4]


class ContactType(IntEnum):
    CHAT = 1
    REPEATER = 2
    ROOM = 3
    SENSOR = 4


class MessageStatus(IntEnum):
    PENDING = auto()
    SENT = auto()
    DELIVERED = auto()
    FAILED = auto()


CONTACT_FLAG_FAVORITE = 0x01
CONTACT_FLAG_TELE_BASE = 0x02
CONTACT_FLAG_TELE_LOC = 0x04
CONTACT_FLAG_TELE_ENV = 0x08


@dataclass
class Contact:
    public_key: bytes
    name: str
    type: int = ContactType.CHAT
    flags: int = 0
    path_length: int = -1  # -1 = flood
    path: bytes = b""
    path_hash_size: int = 1  # 1, 2, or 3 bytes per hop
    force_flood: bool = False
    auto_reset_path: bool = True
    smart_route_selection: bool = False
    latitude: float | None = None
    longitude: float | None = None
    last_seen: datetime | None = None
    received_at: datetime | None = None
    device_lastmod: datetime | None = None
    last_message_at: datetime | None = None
    is_active: bool = True

    @property
    def public_key_hex(self) -> str:
        return self.public_key.hex()

    @property
    def type_label(self) -> str:
        labels = {
            ContactType.CHAT: "Chat",
            ContactType.REPEATER: "Repeater",
            ContactType.ROOM: "Room",
            ContactType.SENSOR: "Sensor",
        }
        return labels.get(self.type, "Unknown")

    @property
    def path_label(self) -> str:
        if self.smart_route_selection:
            return "Smart"
        if self.force_flood:
            return "Flood"
        if self.path_length < 0:
            return "Flood"
        if self.path_length == 0:
            return "Direct"
        return f"{self.path_length} hops"

    @property
    def path_hops(self) -> list[str]:
        """Split path bytes into individual hop prefixes using path_hash_size."""
        if self.path_length <= 0 or not self.path:
            return []
        hs = self.path_hash_size
        return [
            self.path[i : i + hs].hex().upper()
            for i in range(0, min(len(self.path), self.path_length * hs), hs)
        ]

    @property
    def is_favorite(self) -> bool:
        return bool(self.flags & CONTACT_FLAG_FAVORITE)

    @property
    def has_location(self) -> bool:
        if self.latitude is None or self.longitude is None:
            return False
        lat, lon = self.latitude, self.longitude
        if not (-90.0 <= lat <= 90.0 and -180.0 <= lon <= 180.0):
            return False
        # Filter out (0,0)..(1,1) range — unset or deliberately hidden
        return not (0.0 <= lat <= 1.0 and 0.0 <= lon <= 1.0)


@dataclass
class Message:
    sender_key: bytes
    text: str
    timestamp: datetime
    is_outgoing: bool
    is_cli: bool = False
    status: MessageStatus = MessageStatus.PENDING
    message_id: str | None = None
    retry_count: int = 0
    expected_ack_hash: int | None = None
    sent_at: datetime | None = None
    delivered_at: datetime | None = None
    trip_time_ms: int | None = None
    path_length: int | None = None
    path_bytes: bytes = b""
    room_sender_name: str = ""  # resolved sender name for room messages

    @property
    def sender_key_hex(self) -> str:
        return self.sender_key.hex()


class NotificationLevel(IntEnum):
    DEFAULT = -1
    ALL = 0
    MENTIONS = 1
    NONE = 2


@dataclass
class Channel:
    index: int
    name: str
    psk: bytes  # 16 bytes
    unread_count: int = 0
    notification_level: NotificationLevel = NotificationLevel.DEFAULT
    region: str = ""

    PUBLIC_CHANNEL_PSK = "8b3387e9c5cdea6ac9e5edbaa115cd72"

    @property
    def psk_hex(self) -> str:
        return self.psk.hex()

    @property
    def is_empty(self) -> bool:
        return not self.name and all(b == 0 for b in self.psk)

    @property
    def is_public_channel(self) -> bool:
        return self.psk_hex == self.PUBLIC_CHANNEL_PSK

    @property
    def is_hashtag_channel(self) -> bool:
        if self.is_public_channel or not self.name.startswith("#"):
            return False
        expected = hashlib.sha256(self.name.encode()).digest()[:16]
        return self.psk == expected

    @property
    def channel_type(self) -> str:
        if self.is_public_channel:
            return "Public"
        if self.is_hashtag_channel:
            return "Hashtag"
        return "Private"

    @staticmethod
    def empty(index: int) -> "Channel":
        return Channel(index=index, name="", psk=bytes(16))


@dataclass
class DeviceInfo:
    # From SELF_INFO (meshcore reader field names)
    name: str = ""  # Node name
    public_key: str = ""  # Hex string
    tx_power: int = 0
    max_tx_power: int = 0
    radio_freq: float = 0.0  # MHz
    radio_bw: float = 0.0  # kHz
    radio_sf: int = 0
    radio_cr: int = 0
    cr_old_encoding: bool = False
    # From DEVICE_INFO
    model: str = ""  # Board/manufacturer name
    ver: str = ""  # Firmware version string
    fw_build: str = ""  # Build date
    fw_ver: int = 0  # Firmware version code
    max_contacts: int = 32
    max_channels: int = 8
    path_hash_mode: int = 0  # 0=1-byte, 1=2-byte, 2=3-byte (fw v10+)
    multi_acks: int = 0  # 0=off, 1=send extra ACK (fw v10+)
    repeat: bool = False  # client repeat / off-grid mode (fw v9+)
    ble_pin: int = 0  # BLE pairing PIN (from DEVICE_INFO, fw v3+)
    # Location from SELF_INFO (adv_lat/adv_lon)
    latitude: float | None = None
    longitude: float | None = None
    # From BATT_AND_STORAGE
    battery_mv: int = 0
    storage_used_kb: int = 0
    storage_total_kb: int = 0

    @property
    def battery_percent(self) -> int:
        return battery_percent_from_mv(self.battery_mv)

    @property
    def storage_percent(self) -> int:
        if self.storage_total_kb == 0:
            return 0
        return int(self.storage_used_kb * 100 / self.storage_total_kb)


PAYLOAD_TYPENAMES = [
    "REQ",
    "RESPONSE",
    "TEXT_MSG",
    "ACK",
    "ADVERT",
    "GRP_TXT",
    "GRP_DATA",
    "ANON_REQ",
    "PATH",
    "TRACE",
    "MULTIPART",
    "CONTROL",
]

ROUTE_TYPENAMES = ["TC_FLOOD", "FLOOD", "DIRECT", "TC_DIRECT"]


@dataclass
class RxLogEntry:
    timestamp: datetime
    snr: float
    rssi: int
    payload_type: int
    route_type: int
    path: list = field(default_factory=list)
    path_hash_size: int = 1
    pkt_hash: str = ""
    size: int = 0
    transport_from: str = ""
    transport_to: str = ""
    raw_hex: str = ""

    @property
    def payload_type_name(self) -> str:
        if 0 <= self.payload_type < len(PAYLOAD_TYPENAMES):
            return PAYLOAD_TYPENAMES[self.payload_type]
        return f"UNK({self.payload_type})"

    @property
    def route_type_name(self) -> str:
        if 0 <= self.route_type < len(ROUTE_TYPENAMES):
            return ROUTE_TYPENAMES[self.route_type]
        return f"UNK({self.route_type})"


@dataclass
class ChannelMessage:
    channel_index: int
    sender_name: str
    text: str
    timestamp: datetime
    is_outgoing: bool = False
    status: MessageStatus = MessageStatus.DELIVERED
